#include <iostream>
#include <fstream>
#include <cstdlib>
#include "boost/multi_array.hpp"
#include "mm_image.h"
#include "mm_image_fileinput.h"
#include "mm_image_qtoutput.h"
#include "mm_image_funcs.h"
#include "mm_property_image.h"
#include <mm_multi_array_op.h>
#include <algorithm>
#include "mm_canny.h"
#include "tools.h"
#include <vector>
#include <cmath>

using namespace std;
using namespace mimas;
using namespace boost;

//Is there any advantage of using a class instead of a struct?
//I have defined many time a point and a line how to systematize to
//include in mimas/or using a standard library??

struct _todouble: public unary_function< int, double > {
  double operator()( const int &x )
  { return static_cast< double >( x ); }
};

MM_ARRAY_OP( todouble, _todouble )

  typedef multi_array<double,2> Array2D;

/**
Line finding using Hough Transform. A bit
more of work needed to put it in the include of mimas.
Check what can be borrowed from mm_hough_transform

@author: Manuel Boissenin
@date: 28 november 2004
*/

class Hough{
private:
  class Point{
  public:
    int x;
    int y;
    Point(int _x = 0, int _y = 0):
      x(_x),y(_y){}
  };
  
  class Line{
  public:
    Point p1;
    Point p2;
    
    Line(Point &p1_, Point &p2_)
      :p1(p1_),p2(p2_){}
  };
 
  int center_x, center_y;

  //stand for Polar to Cartesian
  //helper function to convert a point
  //from the parametre space to a line in the image space. Add the line
  //to the vector of line.
  void PtoC(double theta,double r)
  {
    int test = 1000;
    //calculate the intersection point between the line and its perpenducular
    //that is passing through the origin
    
    double x = r * cos (double(theta*3.14115/nbOrientationDiv));
    double y = r * sin (double(theta*3.14115/nbOrientationDiv));
    //the direction of the line is (y,-x)
    //scale and check in  the bound of image if necessary

    //we take 2 points on the line centered on intersection,separated by
    //2*test distance
    double size = sqrt(x*x + y*y);

    Point p1 = Point( int(x + y * test /size) + center_x , center_y - int(y - x * test /size) );
    Point p2 = Point( int(x - y * test /size) + center_x , center_y - int(y + x * test /size) );

    lines.push_back(Line(p1,p2));
  }

  //this function has been adapted from Parker's book
  //core of the hough transform: transfrom edge points into parametre space for line 
  void hough (void)
  {
    double conv = 3.1415926535/nbOrientationDiv;
    //Centering the frame on the centre of the image we divide by 2 the sample space needed
    //for r as this value can be negative and its max(|r|) is divided by 2.
    //The two first loop are used all the time -> find a generic way to apply an operation to each pixel of an image.
    for (int i = 0; i < x.getWidth(); i++)
      for (int j = 0; j < x.getHeight(); j++)
	if (x.getPixel(i,j)==255)
	  for (int omega = 0; omega < nbOrientationDiv; omega++) 
	    {
	      int r = (int)( (i - center_x) * cos(double(omega*conv)) - (j - center_y) * sin(double(omega*conv)));
	      
	      assert(imageDiagonalSize / 2 + r > 0 ||imageDiagonalSize / 2 + r <= imageDiagonalSize);
	      parameterSpace[omega][imageDiagonalSize / 2 + r] += 1;
	    }
  }
  
  void findMax(void)
  {
    maxVal  = *(std::max_element(parameterSpace.data(), parameterSpace.data()+parameterSpace.num_elements()));
  }

  void findLine(void){
    //rubish code: use template programing
    for (int t=0; t<nbOrientationDiv; t++)
      for (int r=0; r< imageDiagonalSize; r++)
	if ((parameterSpace[t][r] > maxVal / thresh))
	  {
	    //cerr<< "thresholded value is:" << parameterSpace[t][r] <<  endl;
	    //cerr<< "angle is: " << t <<endl;
	    //cerr <<"dist is  " << r -imageDiagonalSize / 2<<endl;
	    PtoC( t,r - imageDiagonalSize /2);
	  }
  }

  void displayLine(void){
    for (vector<Line>::iterator it = lines.begin(); it != lines.end();it++)
      {
	x.drawLine(it->p1.x, it->p1.y, it->p2.x, it->p2.y, 100);
      }

    //Amazing element-wise multplication developed by Jan
    houghDisplay << multi_array_to_mm_image( parameterSpace * ( 255.0 / maxVal ) )<< mimas::pause();

    houghDisplay << x << mimas::pause();
  }  

public:

  int imageDiagonalSize;
  mm_image_qtoutput<unsigned char> houghDisplay;
    
    
  vector<Line> lines;
  Array2D parameterSpace;
  int nbOrientationDiv;
  double thresh;

  double maxVal;
  mm_image<unsigned char> &x;
  
  
  Hough(mm_image<unsigned char> &image, QApplication &app, double thresh_factor = 2,int orientationDivision = 180)
    :imageDiagonalSize(int(sqrt(double(image.getWidth()*image.getWidth()+image.getHeight()*image.getHeight())))),
     houghDisplay(app),
     parameterSpace(extents[orientationDivision][imageDiagonalSize]), //size has to be checked maybe +1 would be the right measure 
     nbOrientationDiv(orientationDivision),
     thresh(thresh_factor),
     x(image)
  {
    center_x = image.getWidth()/2;
    center_y = image.getHeight()/2;

    //generic programming !!!!! initialisation of the parameterSpace. Is it really necessary?    
    for (int r = 0; r < imageDiagonalSize; r++)
      for (int omega = 0; omega < nbOrientationDiv; omega++)
	parameterSpace[omega][r] = 0;
  }

  
  void exec(void)
  {
    hough();
    findMax();
    findLine();
    displayLine();
  }
};
    
int main(int argc, char **argv)
{
  int retVal = 0;
  try{
    QApplication app( argc, argv );
  
    mm_image<unsigned char> edgeImage;
    mm_image<mm_rgba<unsigned char> > original;
  
    mm_canny<unsigned char> canny;
  
    //canny.setCannyLowThres(20);
    //canny.setCannyHighThres(91)
    canny.setCannyAutoThres(true); // default is true anyway
    canny.setCannyStdDev(1);

    ifstream file( argv[1], ios::binary ); //should send an exception if no filename is present
    //is it my job to do this or should it be embeded with ifstream or >> operator.
    
    file >> original;

    edgeImage = mm_convert_image< unsigned  char > (original);

    canny.canny(edgeImage);

    mm_image_qtoutput< unsigned char  > display( app );
    display << edgeImage << mimas::pause();

    Hough h(edgeImage,app, atof(argv[2]) , atoi(argv[3]));
    h.exec();
  } catch (exception &e) {
    cerr << e.what() << endl;
    retVal = 1;
  };
  return retVal;
}

