/*
 *
 * Multiresolution Chamfer Matching of lenses
 *
 * Bala Amavasai <bala@amavasai.org>
 *
 * $Header: /cvs/mimas/examples/findLenses/main.cc,v 1.13 2005/01/11 16:31:50 engjw Exp $
 */

#include <cmath>
#include <ctime>
#include <fstream>
#include <iostream>
#include <xalan/XPath/XPathInit.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include "mm_image.h"
#include "mm_image_fileinput.h"
#include "mm_image_qtoutput.h"
#include "mm_image_funcs.h"
#include "mm_chamfer_match.h"

using namespace std;
using namespace mimas;

int main(int argc, char **argv)
{
  int retVal = 0;

  XMLPlatformUtils::Initialize();

  try {
   // Static initalisation of XPath utilities.
   XPathInit theXPathInit;

   QApplication app( argc, argv );
	mm_image<int> scene, object, object2, rotobject, original;
	mm_chamfer_match chamferMatch, chamferMatch2;
   mm_image_qtoutput< unsigned char > display( app );

	int xmin, ymin, xmin2, ymin2;
	float minangle, minangle2, corrval, corrval2;
	time_t starttime, endtime;

   MMERROR( argc >=3 && argc <= 4, mm_exception, ,
            "Syntax : " << argv[0] << " <Scene image> "
            "<Object_Parameter_File> [<Object_Parameter_File>]" );

	// load scene image
   ifstream file1( argv[1], ios::binary );
   file1 >> scene;

	original = scene;

	// load object parameter file
	chamferMatch.loadChamferParameters(argv[2]);	
	
	// load object image
	ifstream file2( (chamferMatch.getName()).c_str(), ios::binary);
	file2 >> object;
		
	// load second object image 
	if (argc == 4) {
		chamferMatch2.loadChamferParameters(argv[3]);
		ifstream file3( (chamferMatch2.getName()).c_str(), ios::binary );
		file3 >> object2;		
	}
	
	cout << "Searching for object..." << endl;
	 
	starttime=time(NULL);

	// add scene to chamferMatch objects
	chamferMatch.addScene(scene);
	if (argc == 4) chamferMatch2.addScene(scene);

	// Search for objects and return xmin, ymin and angle
	chamferMatch.findModel(object, xmin, ymin, minangle);
	corrval = chamferMatch.getMinCorrVal();
	
	if (argc == 4) {
		chamferMatch2.findModel(object2, xmin2, ymin2, minangle2);
		corrval2 = chamferMatch2.getMinCorrVal();
	}

	endtime=time(NULL);

	cout << "Object(s) found... Time = " << endtime-starttime << " seconds" << endl;


	// Draw X at location where object(s) is(are) found
	original.drawLine(xmin - object.getWidth()/2, ymin - object.getHeight()/2, 
					  xmin + object.getWidth()/2, ymin + object.getHeight()/2, 191);
	original.drawLine(xmin + object.getWidth()/2, ymin - object.getHeight()/2, 
					  xmin - object.getWidth()/2, ymin + object.getHeight()/2, 191);
	
	if (argc == 4) {
		original.drawLine(xmin2 - object2.getWidth()/2, ymin2 - object2.getHeight()/2,
						  xmin2 + object2.getWidth()/2, ymin2 + object2.getHeight()/2, 191);
		original.drawLine(xmin2 + object2.getWidth()/2, ymin2 - object2.getHeight()/2,
						  xmin2 - object2.getWidth()/2, ymin2 + object2.getHeight()/2, 191);
	}

	cout << "(x,y) = " << xmin << "," << ymin << "   angle = " << minangle*180.0/M_PI <<
		" degrees corrval = " << corrval << " " << endl;

	if (argc == 4) { 
		cout << "(x2,y2) = " << xmin2 << "," << ymin2 << "   angle2 = " << minangle2*180.0/M_PI <<
			" degrees corrval2 = " << corrval2 << " " << endl;
	}

	display << original << mimas::pause();
  } catch ( exception &e ) {
    cerr << e.what() << endl;
    retVal = 1;
  };

  XMLPlatformUtils::Terminate();

  return retVal;
}

