// stumeikle.
// Mon 21 Oct 2002 19:43:09 CEST
// playing with alternatives to pairwise
// this version: using a pairwise approach to forming a location hough
// but without the histograms
//

//
//
// Syntax : findTux <scene image>
//
//
#include <boost/numeric/ublas/vector.hpp>
#include <cassert>
#include <fstream>
#include <qapplication.h>
#include <qpainter.h>
#include "findTuxWidget.h"
#include "mm_linalg.h"
#include "mm_pol_model.h"
#include "mm_pol_scene.h"
#include "mm_image_fileinput.h"

using namespace boost;
using namespace std;
using namespace mimas;

typedef boost::numeric::ublas::vector< double > Vector;

FindTuxWidget::FindTuxWidget( QWidget *parent, const char *name, WFlags f ):
  QWidget( parent, name, f )
{
  MMERROR( qApp->argc() == 2, mm_exception, ,
           "Usage: " << qApp->argv()[0] << " imagefile" << endl
           << "       searches for penguin_model in imagefile" );

  cout << "Creating all new model..." << endl;

  mm_pol_model p;
    
  //switch on scale invariance AT THE MODEL...
  p.setScaleInvariance(true);
  p.createFromImageFile("../images/stills/penguin_model.ppm");

  ml = p.getModelLines();

  cout << "...Finished." << endl;

  mm_pol_scene		s;
  
  ifstream imfile( qApp->argv()[1], ios::binary );
  imfile >> im;

  s.createFromImage(im);
  cout << "Locating model in scene..." << endl;
  matches = s.getMatches(p);

  cout << "Finished searching.. did i find it?" << endl;
}

FindTuxWidget::~FindTuxWidget(void)
{
  /*
  //delete the match info
  for( mm_pol_match_ptr_list::iterator mi=matches.begin();
       mi!=matches.end(); mi++ )
    delete (*mi);
  matches.clear();
  */
}

void FindTuxWidget::paintEvent( QPaintEvent * )
{
  QPainter paint( this );

  //draw the current input scene 
  //draw the bit map
  QImage img(im.getWidth(),im.getHeight(),32);
  
  for(int y=0;y<im.getHeight();y++ )
    for(int x=0;x<im.getWidth();x++ )
      {
        int	p = im.getPixel(x,y);
        img.setPixel( x,y, qRgb(p,p,p));
      }
  
  QPixmap pix;
  pix.convertFromImage(img);
  
  bitBlt( this , 0,0, &pix, 0,0, im.getWidth(), im.getHeight(), CopyROP,
          false);

  for( mm_pol_match_ptr_list::iterator it=matches.begin();
       it!=matches.end(); it++ )
    {
      mm_pol_match_ptr pm=(*it);
	    
      Vector d( pm->getPosition() );
      mm_angle	a = (double)pm->getOrientation();
      double		s = pm->getScale();

      paint.setPen(Qt::red);

      //try and draw them properly now
      mm_line_ptr_list::iterator	lineit;
      for( lineit=ml.begin(); lineit!=ml.end(); lineit++)
        {
          mm_line_ptr l =(*lineit);
          Vector start(2),end(2);

          // For some reason 'rotate( l->getStart() * s, a )' won't compile in
          // release-mode.
          start = l->getStart() * s;
          end   = l->getEnd()   * s;

          start = rotate( start, -a ) + d;
          end   = rotate( end  , -a ) + d;
		    
          paint.drawLine( (int)start(0), (int)start(1),
                          (int)end(0), (int)end(1) );
        }
    }
}

