//
// Filter by colour - brainstorming
// Stuart Meikle
// Mon Aug 23 14:37:05 UTC 2004
// $Header: /cvs/mimas/examples/hue/main.cc,v 1.5 2004/12/06 18:35:24 engjw Exp $
//

#include <fstream>
#include <iostream>
#include "mm_image_fileinput.h"
#include "mm_image_fileoutput.h"
#include "mm_image_qtoutput.h"
#include <qapplication.h>
#include "mm_property_image.h"
#include "mm_rgba.h"
#include "mm_edge_string.h"
#include "mm_edge_string_op.h"
#include "mm_canny_tool.h"

using namespace mimas;
using namespace std;

int main(int argc, char **argv)
{
  int retVal = 0;
  try {
    QApplication app( argc, argv );

    MMERROR( argc >= 3, mm_exception, ,
             "Please see the README file for details of how to run this app" );
    
    cout << endl;
    cout << "Matching primary hues extracted from the model to the scene..." << endl; 
    cout << "(Could be used as a simple locate-object algorithm. Or used for filtering (see findTux_byHue)" << endl;
    cout << endl;
    
    // load in the model ...
    mm_image< mm_rgba< int > > im;
    ifstream imFile( argv[1], ios::binary ); imFile >> im;

    mm_image< mm_rgba< int > > sim;

    ifstream simFile( argv[2], ios::binary ); simFile >> sim;

    //create a new image which just contains the hue values
    mm_image<int>		oim;

    oim.init( im.getWidth(), im.getHeight() );
    int	x,y,val;
    double	hist[360],smooth_hist[360];
    for(x=0;x<360;++x)	hist[x]=0;

    for(y=0;y<im.getHeight();++y)
    {
    	for (x=0;x<im.getWidth();++x)
	{
	    double	r,g,b;

       mm_rgba< int > *pixel = &im.getPixel( x, y );
	    r = (double)pixel->getRed();
	    g = (double)pixel->getGreen();
	    b = (double)pixel->getBlue();

	    if (r==0.0 && g==249.0 && b==162.0) //our transparency colour
	    	continue;

		//cout << "Pixel hue=" << (int)im.getPixelHue(x,y) << endl;
	    hist[(int)pixel->getHue()] ++;
	}
    }

    //first get rid of spikes... and find max
    double	maxval=0;
    smooth_hist[0]=hist[0];
    smooth_hist[359]=hist[359];
    for(x=1;x<359;x++)
    {
 //       smooth_hist[x]=(hist[x-1]+hist[x+1])/2.0;
	smooth_hist[x]=hist[x];
	if (hist[x]>maxval)
	    maxval=hist[x];
    }

    //threshold
    for (x=0;x<360;++x)
    {
        if (smooth_hist[x]<(maxval*0.2))
	    smooth_hist[x]=0;
    }

    mm_image<int>	temp;
    temp.init( 360, 100 );

    //find the peaks
    for(x=0;x<360;x++)
    {
        hist[x]=0;
        if (x==0)
	{
	    if (smooth_hist[0]>smooth_hist[1])
	    {
	        hist[0]=smooth_hist[0];
		continue;
	    }
	}
	if (x==359)
	{
	    if (smooth_hist[359]>smooth_hist[358])
	    {
	        hist[359]=smooth_hist[359];
		continue;
	    }
	}

	if (smooth_hist[x]>smooth_hist[x-1] && smooth_hist[x]>smooth_hist[x+1])
	    hist[x]=smooth_hist[x];
    }

    for(x=0;x<360;x++)
    	temp.drawLine( x,99,x, 99-(int)(hist[x]*100.0/maxval), 255);
    //temp.display();

    // now load the new scene
    mm_image< mm_rgba< int > > soim;
    soim.init(sim.getWidth(),sim.getHeight());

    for(y=0;y<sim.getHeight();++y)
    {
    	for (x=0;x<sim.getWidth();++x)
	{
     val = (int) sim.getPixel( x, y ).getHue();

	    if (val>=0 && val<=359)
	    {
	    	//if (hist[val]==0)
		//	sim.setPixel(x,y,0);
		if (hist[val]!=0)
		{
			int a,b;
			for(a=-2;a<3;++a)
				for(b=-2;b<3;++b)
				{
              soim.setPixel( x + b, y + a, sim.getPixel( x + b, y + a ) );
				}
		}
	    }
	}
    }
    
    cout << "Finished..." << endl;

    mm_image_qtoutput< mm_rgba< unsigned char > > display( app );
    display << soim << mimas::pause();
  } catch ( exception &e ) {
    cerr << e.what() << endl;
    retVal = 1;
  };
  return retVal;
}
