/*
 * $Header: /cvs/mimas/examples/multiresChamferMatch/main.cc,v 1.13 2004/12/09 23:27:18 engjw Exp $
 *
 * Multiresolution Chamfer Matching
 * Notes: algorithms are in mm_image_funcs.h
 *        requires edges from mm_canny.h
 */

#include <cassert>
#include <iostream>
#include <cmath>
#include <ctime>
#include <fstream>
#include "mm_image.h"
#include "mm_image_fileinput.h"
#include "mm_image_qtoutput.h"
#include "mm_image_funcs.h"
#include "mm_canny.h"

using namespace std;
using namespace mimas;

int main(int argc, char **argv)
{
  int retVal = 0;
  try {
   QApplication app( argc, argv );
   mm_image_qtoutput< unsigned char > display( app );
	mm_image<int> inimage, object, rotobject, original;
	mm_canny<int> canny;

	time_t starttime;

   ifstream inimagefile( "images/scene-rotate-20.pgm", ios::binary );
   inimagefile >> inimage;
   ifstream objectfile( "images/rgripper.pgm", ios::binary );
   objectfile >> object;

	canny.setCannyAutoThres(false);
	canny.setCannyStdDev(1.0);
	canny.setCannyLowThres(11);
	canny.setCannyHighThres(91);

   original = inimage;

	canny.canny(inimage);
	canny.canny(object);

   display << inimage << mimas::pause();
   display << object << mimas::pause();

	int x,y, xmin = INT_MIN, ymin = INT_MIN;
   assert( ( -FLT_MAX ) < 0.0 );
	float retcorr, mincorr, minangle = -FLT_MAX, angle;

	mincorr=MAXFLOAT;
	starttime=time(NULL);
	cerr << "Searching for object";
	for (angle=-30.0; angle<=30.0; angle+=5.0)
	{
		paethRotateRadian(object,rotobject,M_PI*angle/180.0);

		multiLevelCorrelation(inimage, rotobject, 3, x, y,
		                      retcorr,true);
		if (retcorr<mincorr)
		{
			mincorr=retcorr;
			xmin=x; ymin=y;
			minangle=angle;
		}


	}

	cout << "...object found (I think!). Time = " << time(NULL)-starttime << " seconds"
	<< endl;

   assert( xmin != INT_MIN && ymin != INT_MIN && minangle != -FLT_MAX );

	// Draw X at location where object is found
	original.drawLine(xmin,ymin,xmin+object.getWidth(),ymin+object.getHeight(),0);
	original.drawLine(xmin+object.getWidth(),ymin,xmin,ymin+object.getHeight(),0);

	// Overlay object onto original image
	paethRotateRadian(object, rotobject, M_PI*minangle/180.0);
	for (int j=0; j<rotobject.getHeight(); j++)
		for (int i=0; i<rotobject.getWidth(); i++)
			if (rotobject.getPixel(i,j)>127)
				original.setPixel(i+xmin,j+ymin,255);


	cout << "(x,y) = " << xmin << "," << ymin << "   angle = " << minangle << " corrval = " << mincorr << " "
	<< endl;

   display << original << mimas::pause();
  } catch ( exception &e ) {
    cerr << e.what() << endl;
    retVal = 1;
  };
  return retVal;
}

