#ifndef __LSE_H
#define __LSE_H

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/multi_array.hpp>
extern "C" {
#include <f2c.h>
#include <clapack.h>
}
#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif



template< typename T >
struct getrf_
{
  int operator()( integer *m, integer *n, T *a, integer *lda, integer *ipiv,
                  integer *info );
};

template< typename T >
struct getri_
{
  int operator()( integer *n, T *a, integer *lda, integer *ipiv, T *work,
                  integer *lwork, integer *info );
};

template< typename T >
struct mm_inv
{
  boost::numeric::ublas::matrix< T, boost::numeric::ublas::column_major >
    operator()( const boost::numeric::ublas::matrix< T,
                boost::numeric::ublas::column_major > &m ) {
    assert( m.size1() == m.size2() );
    boost::numeric::ublas::matrix< T, boost::numeric::ublas::column_major >
      retVal( m );
    integer
      N = m.size1(),
      LDA = m.size1(),
      LWORK = -1,
      IPIV[N],
      INFO;
    getrf_< T >()( &N, &N, retVal.data().begin(), &LDA, &IPIV[0], &INFO );
    assert( INFO == 0 );
    T DLWORK;
    getri_< T >()( &N, retVal.data().begin(), &LDA, &IPIV[0], &DLWORK, &LWORK,
                  &INFO );
    LWORK = (integer)DLWORK;
    T WORK[LWORK];
    getri_< T >()( &N, retVal.data().begin(), &LDA, &IPIV[0], &WORK[0],
                   &LWORK, &INFO );
    return retVal;
  }
};

template< typename T >
struct gels_
{
  int operator()( char *trans, integer *m, integer *n, integer *nrhs,
                  T *a, integer *lda, T *b, integer *ldb, T *work,
                  integer *lwork, integer *info );
};

template< typename T >
struct mm_gels
{
  boost::numeric::ublas::vector< T > operator()
       ( const boost::numeric::ublas::matrix< T,
         boost::numeric::ublas::column_major > &_A,
         const boost::numeric::ublas::vector< T > &_b ) {
    boost::numeric::ublas::matrix< T,
      boost::numeric::ublas::column_major > A( _A );
    boost::numeric::ublas::vector< T > b( _b );
    assert( A.size1() >= A.size2() );
    assert( A.size1() == b.size() );
    char TRANS = 'N';
    integer
      M = A.size1(),
      N = A.size2(),
      NHRS = 1,
      LDA = A.size1(),
      LDB = b.size(),
      LWORK = -1,
      INFO;
    T DLWORK;
    gels_< T >()( &TRANS, &M, &N, &NHRS, A.data().begin(), &LDA,
                  b.data().begin(), &LDB, &DLWORK, &LWORK, &INFO );
    assert( INFO == 0 );
    LWORK = (integer)DLWORK;
    T WORK[LWORK];
    gels_< T >()( &TRANS, &M, &N, &NHRS, A.data().begin(), &LDA,
                  b.data().begin(), &LDB, &WORK[0], &LWORK, &INFO );
    assert( INFO == 0 );
    boost::numeric::ublas::vector< T > retVal( A.size2() );
    boost::copy_n( b.data().begin(), retVal.size(), retVal.data().begin() );
    return retVal;
  }
};


#endif
