//
// Example application that uses rotationChamferCorrelationEdgeMap
// Bala Amavasai <bala@amavasai.org>
//
// $Header: /cvs/mimas/examples/rotateChamferMatch/main.cc,v 1.10 2004/12/09 23:27:19 engjw Exp $
//


#include <iostream>
#include <cmath>
#include <ctime>
#include <fstream>
#include "mm_image.h"
#include "mm_image_fileinput.h"
#include "mm_image_qtoutput.h"
#include "mm_image_funcs.h"
#include "mm_canny.h"

using namespace std;
using namespace mimas;

int main(int argc, char **argv)
{
  int retVal = 0;
  try {
   QApplication app( argc, argv );
   mm_image_qtoutput< unsigned char > display( app );
	mm_image<int> inimage(255), object(255), originalobject(0), rotoriginalobject(0), tempobject(255), original(255);
	mm_canny<int> canny;

	time_t starttime;// , endtime;

   ifstream inimagefile( "firstSeq0063.pgm", ios::binary );
   inimagefile >> inimage;
   original = inimage;
   ifstream objectfile( "small-lens.pgm", ios::binary );
   objectfile >> object;
   originalobject = object;

//	canny.setCannyAutoThre0(false);
	canny.setCannyStdDev(1.4);
//	canny.setCannyLowThres(1);
//	canny.setCannyHighThres(90);


	// Edge detect and perform search on lower resolution
	canny.canny(inimage);
	halfResolutionEdgePyramid(inimage);
   display << inimage << mimas::pause();

	canny.canny(object);
	halfResolutionEdgePyramid(object);
   display << object << mimas::pause();

   starttime=time(NULL);
	cerr << "Searching for object";

	int xmin, ymin, minangle;
	float mincorr;
	int resfactor=2; // resolution multiply factor e.g. if halfResolution is run twice than this value is 4

	// Do Chamfer matching
	rotationChamferCorrelationEdgeMap(inimage,object,0,359,10,xmin,ymin,minangle,mincorr);

	// Output some information	
	cout << "...object found (I think!). Time = " << time(NULL)-starttime << " seconds"
	    << endl; 
	cout << "(x,y)=" << xmin << "," << ymin << "   angle=" << minangle << " corrval=" << mincorr << " " << endl;

	
	// Draw an X at the point that object is found
	original.drawLine(xmin*resfactor,ymin*resfactor,resfactor*(xmin+object.getWidth()),resfactor*(ymin+object.getHeight()),63);
	original.drawLine(resfactor*(xmin+object.getWidth()),resfactor*ymin,resfactor*xmin,resfactor*(ymin+object.getHeight()),63);


	// superimpose object on scene
	canny.canny(originalobject);
	paethRotateRadian(originalobject,rotoriginalobject, (float)minangle*M_PI/180.0);
	for (int j=0; j<originalobject.getHeight(); j++)
	   for (int i=0; i<originalobject.getWidth(); i++)
	       if (rotoriginalobject.getPixel(i,j)==255)
		   original.setPixel(resfactor*xmin+i,resfactor*ymin+j,255);

   display << original << mimas::pause();
  } catch ( exception &e ) {
    cerr << e.what() << endl;
    retVal = 1;
  };
  return retVal;
}

