//
// Binary image morphology tool
//


#ifndef MM_MORPHOLOGY_TOOL_H
#define MM_MORPHOLOGY_TOOL_H

#include <iostream>
#include <cmath>
#include <values.h>
#include "mm_image.h"
#include "mm_rgba.h"

namespace mimas {


/**
 * Morphology tool for binary images
 * @author Bala Amavasai (bala@amavasai.org)
 * @date Tue Feb 18 16:27:41 GMT 2003
 * $Header: /cvs/mimas/include/mm_binmorph_tool.h,v 1.11 2004/12/09 23:27:19 engjw Exp $
 */


class mm_binmorph_tool //: public mm_object
{
	private:
		mm_image<int> structElement; ///< the structuring element
		enum operation {mm_binmorph_dilation, mm_binmorph_erosion};
		void morphology(mm_image<int> &imagein, mm_image<int> &imageout, mm_binmorph_tool::operation op );
		int structElementOnes(void); ///< the number of 1's in the structuring element
		void labelConnectedGroup(mm_image<int> &imagein, mm_image<int> &imageout, int x, int y, int label); // private recursive function for connected components

	public:
		mm_binmorph_tool(void); ///< constructor
		~mm_binmorph_tool(void); ///< destructor
		// void setStructElement(mm_image<int> &structElem); ///< redefine the structuring element
		void greyToBinaryOtsu(mm_image<int> &imagein, mm_image<int> &imageout); ///< convert imagein to binary form based on Otsu's optimal method and store in imageout
		void greyToBinary(mm_image<int> &imagein, mm_image<int> &imageout, int threslevel); ///< convert imagein to binary form based on user defined threshold level and store in imageout
		void greyToBinary(mm_image<int> &imagein, int threslevel); ///< convert imagein to binary form based on user defined threshold level
		void binaryToGrey(mm_image<int> &imagein); ///< normalise binary image to the range 0..255
		void normaliseToGrey(mm_image<int> &imagein); ///< normalise imagein to the range 0..255
		void normaliseToColour(mm_image<int> &imagein, mm_image< mm_rgba< int > > &imageout); ///< normalise imagein to colour in the range 0..255*255*255. Useful for displaying connected components
		void negate(mm_image<int> &imagein); ///< negate image so component appears as 1 and background as 0
		void erosion(mm_image<int> &imagein, mm_image<int> &imageout); ///< erode binary image
		void dilation(mm_image<int> &imagein, mm_image<int> &imageout); ///< dilate binary image
		void closing(mm_image<int> &imagein, mm_image<int> &imageout); ///< closing operation on binary image
		void opening(mm_image<int> &imagein, mm_image<int> &imageout); ///< opening operation on binary image
		void difference(mm_image<int> &imagein1, mm_image<int> &imagein2, mm_image<int> &imageout); ///< set of binary points that belong to imagein1 but not on imagein2 *Note that this is not image subtraction!
		void boundaryExtract(mm_image<int> &imagein, mm_image<int> &imageout); ///< set of points on the boundary of set imagein
		void connectedComponents(mm_image<int> &imagein, mm_image<int> &imageout); ///< perform connected components analysis on imagein and store labels into imageout. Use normaliseToGrey to before displaying imageout.
		void distanceTransform(mm_image<int> &imagein, mm_image<int> &imageout); ///< find the distance transform of imagein and store it in imageout. Ref: David Eberly NNTP message id <2vkc27$mjs@trochlea.cs.unc.edu>

		// supplementary functions
		void floodFill(mm_image<int> &imagein, int x, int y, int newval, int oldval);; // flood fill. Replace oldval with newval starting from point x,y
};

};

#endif


