//
// canny stuff. rewritten, though based on tina
// stumeikle Sat May  4 19:22:36 2002
//
//
//




#ifndef MM_FILTER_CANNY_H
#define MM_FILTER_CANNY_H

#include <cmath>
#include "mm_image.h"
#include "mm_gauss_tool.h"
#include "mm_property_image.h"
#include "mm_edge_string.h"
#include "mm_edge_string_op.h"

namespace mimas {
	
/**
 *
 * Canny edge extraction tool. Algorithms and inspiration from TINA.
 * MM_Canny_Tool is an operator which is used to extract edge strings from
 * input images. Edge extraction method is based upon Canny's algorithm and
 * includes Canny's post processing methods. Found edges are grouped into
 * continuous strings. This class is a template class because it must operator
 * on a template class (the input image).
 *
 * @author Stuart Meikle (stu@stumeikle.org)
 * @date Sat May  4 19:22:36 2002
 * $Header: /cvs/mimas/include/mm_canny_tool.h,v 1.10 2004/12/14 23:27:07 engjw Exp $
 */


template<typename T>
class	mm_canny_tool
{
    private:
	double	sigma;
	double	precision;
	double	low_edge_thres;
	double	up_edge_thres;
	int	len_thres;

    //operator. operates on images
    public:

    mm_canny_tool(void)		///< constructor. default prams should be ok for most tasks.
    {
	sigma = 1.0;
	precision = 0.01;
	low_edge_thres = 3.0;
	up_edge_thres = 5.0;
	len_thres = 5;
    }

    mm_edge_string_ptr_list findEdgeStrings( mm_image<T> &input ) ///< find the edges in the given input image and return as a list of edge strings
    {
	//perhaps these first should be of fixed (double?) type
	mm_image<int>		gauss_im, grad_x, grad_y;
	mm_image<int>		gradsq;
	mm_gauss_tool<int>	gauss( 1.0, 0.01 );

	//(1) first gauss the image
	gauss.filter( input, gauss_im );

	//(2) get the horizontal and vertical gradients
	//    grad filter is a simple static function
	//    might be an idea to switch the gauss image to a double first?
	mm_filter<int>::gradX( gauss_im, grad_x );
	mm_filter<int>::gradY( gauss_im, grad_y );

	//(2.5) to save memory, get rid of gauss_im
	gauss_im.clear();

	//(3.0) create grad sq. this is very elegant, buuut very memory intensive
	//gradsq = (grad_x*grad_x) + (grad_y*grad_y);
	//(... so we wont ...)
	mm_filter<int>::sumSquares( grad_x, grad_y, gradsq );

	//(4.0) do non maxima suppression
	// (errr.. shouldnt we look for zero crossings of grad^2 im?)
	mm_edge_image	edge_im;
	mm_filter<int>::nonMaximaSuppression( grad_x, grad_y, gradsq,
					      3.0, edge_im );
	grad_x.clear();
	grad_y.clear();
	gradsq.clear();

	//(5.0) post process the resulting edgels
	// what's the best way to do this ? we could create an mm_edge_image function
	// but it would be better *not* to have specific edge images. better to have image properties
	// than specific image types. anyway mm_image<int> is a perfectly valid edge image too.
	mm_edge_string_ptr_list edge_strings;
	mm_edge_string_op	eop;

	edge_strings = eop.extractFromImage( edge_im, 5, 5.0 );

	//done canny ----------------------------------------------------
	return edge_strings;
    }
};

};

#endif
