//
// distributable pairwise
// stuart meikle
// begun Tue Feb 22 10:45:47 2000
//

// NOTE: its just about possible to split this into a .cc and .h file with
// a modern compiler. YOu might need to use the #template / #interface and #implementation
// directives though. (template hassles)

// common base classes etc for the packages to be distributed
// i'll use mm for miniman to differential this code from that of tina

// i'll be so impressed if corba can send a char* array to a remote process
// NOTE: an image is just a matrix of 'T' type objects. I'm not going to attempt
// to build in palette information here -> that would be the job of who-ever/whatever
// interprets and renders the image, not a property of the image itself.

// image could of course be a vector .... but i'll leave the distinction in for the
// time being

// Thu May  4 14:05:13 2000. stu the meikle
// added Iiiiiiiiiiimage Maaaaaagick stuff today. (Its good so you have to say
// like you're introducing it on stage). enables images to be written to file in
// numerous formats.

// Tue Sep 26 13:56:01 - Bala Amavasai
// added readFromFile - read image from file
//			rawData - pointer to the data - useful for directly thunking data in KLT

// Wed Mar 7 19:06:02 - Bala Amavasai
// ImageMagick functions are all F-ed up since versions from 5.2.9 are 16-bit.
// readFromFile seems okay but writeToFile doesn't want to work
// added    readFromFilePGM - read PGM file
//			   writeToFilePGM  - write PGM to file
// the above two functions require netpbm to be installed - note that some buggy installations
// do not include shopts, in which case compile it 'yerself and install it!

// Sat May  4 14:24:14 2002 - Stu Meikle
// some tidying up. made the destructor virtual, which is more useful for images which
// have pointers at each pixel. (edgel images etc).
// addded a switch for objectID, because it's a big structure! (remember histograms
// are images but often not much bigger than 255 bytes).

// SUGGESTION: keep all QT specific code in the mm_image files only. 
// (in case someone needs to remove / switch it in the future)
//
// $Header: /cvs/mimas/include/mm_image.h,v 1.46 2005/02/24 19:48:45 engjw Exp $
//

#ifndef MM_IMAGE_H
#define MM_IMAGE_H

#include <boost/multi_array.hpp>
#include <algorithm>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <string>

//YIKES QT DEPENDENCIES
#include <qapplication.h>
#include <qlabel.h>
#include <qimage.h>
#include <qpixmap.h>

#include <mm_pixel_convert.hh>
#include "mm_object.h"
#include "mm_exception.h"

#define sgn(x) ((x<0)?-1:((x>0)?1:0))

namespace mimas {

//we could have:

/** Template class for images.
    Hmmm. Well. MM_image is a generic image class. You specify the pixel type
    at creation time. So if you want an image of ints do
    \code mm_image<int> my_image;\endcode
    and so on. Because its a template class we have a few problems. With
    slightly older compilers it was impossible to separate source and headers
    for template classes because the code was modified at compile time to deal
    with all the different uses of the class. Hence i've put all the template
    code in the <em>mm_image.h</em> header. With more modern compilers I think
    that you can have a normal .h and .cc file for templates, and they're
    sorted out ok, but i haven't checked it with gcc 2952 yet. The image class
    is fairly straightwforwards - its really a bit like a big vector (any C++
    gurus might want to make it into a child of the vector class). There are a
    few ops provided such as multiply, scale, threshold, etc...
    
    to display your image (greyscale) in a window. This also hangs the process
    till the window is
    closed. There's also \c ::writetofile to write the data out. */
template<typename T>
class	mm_image : 	public mm_object
			//perhaps public virtual drawable one day
{

	private:
#ifdef DEBUG
		char 	objectID[255]; 
#endif
	protected:
      typedef std::istream istream;
      typedef std::ostream ostream;
		    	//(co-ordinate system can go in here one day)
    	/** Pointer to data.
          Don't be scared. Its just a ptr to im[][] really. */
      boost::multi_array< T, 2 > data;
      // Where we store out of bounds entries.
      // T	outofbounds;
      /** Initial background gray value for image.
          Useful for deciding background color for operations like
          rotations. */
      T 	initVal; 
      /// Dummy-pixel to work on, if initVal is requested.
      T  dummy;

 public:

		int getWidth( void ) const
		{
        return data.shape()[1];
		}
		int getHeight( void ) const
		{
        return data.shape()[0];
		}

		mm_image( void )
		{
        initVal = T();
#ifdef DEBUG			
			sprintf(objectID,"Undeclared");
#endif
		}

		mm_image( T aInitVal ): initVal( aInitVal )
		{
#ifdef DEBUG
			sprintf(objectID,"Undeclared");
#endif
		}

      /// Copy-constructor.
      mm_image( const mm_image<T> &_image )
      {
        operator=( _image );
      }

#ifdef DEBUG
		void setObjectID( char *fn )
		{
			sprintf(objectID,"%s",fn);
		}
#endif	
		virtual void setInitVal( T i )
		{
			initVal=i;
		}


		T getInitVal( void )
		{
			return initVal;
		}

		virtual void init( int w , int h )
		{
        // Only resize, if necessary.
        if ( w != getWidth() || h != getHeight() ) {
          data.resize( boost::extents[ 0 ][ 0 ] );
          data.resize( boost::extents[ h ][ w ] );
        };
        std::fill( data.data(), data.data() + data.num_elements(),
                   initVal );
		}

		virtual void clear( void )
		{
        data.resize( boost::extents[ 0 ][ 0 ] );
		}

      /** Get const-reference to a pixel.
          This method includes clipping! Specifying out-of-image coordinates
          will result in a const-reference to a dummy pixel.
          @param x x-coordinate
          @param y y-coordinate
          @see pixel */
		const T &getPixel( int x, int y ) const
		{
         if ( x>=0 && y>=0 && x<getWidth() && y<getHeight() )
				return data[ y ][ x ];
			else
				return initVal;
		}

      /** Get reference to a pixel.
          This method includes clipping! Specifying out-of-image coordinates
          will result in a reference to a dummy pixel.
          @param x x-coordinate
          @param y y-coordinate
          @see pixel */
		T &getPixel( int x, int y )
		{
         if ( x>=0 && y>=0 && x<getWidth() && y<getHeight() )
           return data[ y ][ x ];
			else {
			  dummy = initVal;
			  return dummy;
			};
		}

      /** Set value of a pixel.
          This method includes clipping!
          @param x x-coordinate
          @param y y-coordinate
          @see pixel */
		virtual void setPixel( int x, int y, T val )
		{
        if ( x>=0 && y>=0 && x<getWidth() && y<getHeight() )
          data[ y ][ x ]  =  val;
		}

      /** Get const-reference to a pixel.
          No clipping!
          @param x x-coordinate
          @param y y-coordinate
          @see getPixel */
		const T& pixel(  int x, int y ) const
		{
         assert( x>=0 && y>=0 && x<getWidth() && y<getHeight() );
			return data[ y ][ x ];
		}

      /** Get reference to a pixel.
          No clipping!
          @param x x-coordinate
          @param y y-coordinate
          @see setPixel */
		T& pixel(  int x, int y )
		{
         assert( x>=0 && y>=0 && x<getWidth() && y<getHeight() );
			return data[ y ][ x ];
		}

      /// Access to aggregated multi-array.
      boost::multi_array< T, 2 > &rawData( void )
		{
			return data;
		}

      /// Read-only access to aggregated multi-array.
		const boost::multi_array< T, 2 > &rawData( void ) const
		{
        return data;
		}

		bool initialised( void ) const
		{
         return data.num_elements() > 0;
		}

		void multiply( T v )
		{
			int	x,y;

			for( y=0;y<getHeight( );y++ )
				for( x=0;x<getWidth( );x++ )
				{
              pixel( x,y ) *= v;
				}
		}

		//some common features here with mm_vector class ...
		//a cleverer man could probably use the same underlying base class
		//but for now it doesnt matter
		double dotProduct( const mm_image<T> &im2 ) const throw (mm_exception)
		{
         MMERROR( im2.getWidth() == getWidth() &&
                  im2.getHeight() == getHeight(),
                  mm_exception, ,
                  "Can't (won't) compute the dot product of two images of"
                  " differing sizes." );

			double result = 0.0;
         const T
           *p = data.data(),
           *q = im2.rawData().data();

         for ( int i=0; i<(signed)data.num_elements(); i++ )
           result += *p++ * *q++;

			return result;
		}

		void scale( T val1 ,T val2 ) throw (mm_exception)
		{
        if ( data.num_elements() > 0 ) {
          T
            minval = *min_element( data.data(),
                                   data.data() + data.num_elements() ),
            maxval = *max_element( data.data(),
                                   data.data() + data.num_elements() );
          MMERROR( maxval > minval, mm_exception, ,
                   "mm_image::scale: Cannot normalize plain image." );
          double factor = (double)( val2 - val1 ) / ( maxval - minval );
          T *p = data.data();
          for( int i=0; i<(signed)data.num_elements(); i++ ) {
            *p = (T)( ( *p - minval ) * factor + val1 );
            p++;
          };
        };
      }

		void threshold( T v )
		{
        T *p = data.data();
        for ( int i=0; i<(signed)data.num_elements(); i++ ) {
          if ( *p < v ) *p = T();
          p++;
        };
		}


		T getMaxVal( void ) const
		{
        T retVal;
        if ( data.num_elements() > 0 )
          retVal = *max_element( data.data(),
                                 data.data() + data.num_elements() );
        else
          retVal = T();
        return retVal;
		}
		
		
		/* Bresenham line function */
		void drawLine(int x1, int y1, int x2, int y2, T colour)
		{
			int i,dx,dy,sdx,sdy,dxabs,dyabs,x,y,px,py;

			dx=x2-x1;			/* the horizontal distance of the line */
			dy=y2-y1;			/* the vertical distance of the line */
			dxabs=abs(dx);
			dyabs=abs(dy);
			sdx=sgn(dx);
			sdy=sgn(dy);
			x=dyabs>>1;
			y=dxabs>>1;
			px=x1;
			py=y1;

			setPixel(px,py,colour);

			if (dxabs>=dyabs) /* the line is more horizontal than vertical */
			{
				for(i=0;i<dxabs;i++)
				{
					y+=dyabs;
					if (y>=dxabs)
					{
						y-=dxabs;
						py+=sdy;
					}
					px+=sdx;
					setPixel(px,py,colour);
				}
			}
			else /* the line is more vertical than horizontal */
			{
				for(i=0;i<dyabs;i++)
				{
					x+=dxabs;
					if (x>=dyabs)
					{
						x-=dyabs;
						px+=sdx;
					}
					py+=sdy;
					setPixel(px,py,colour);


				}
			}
		}
		

		void drawBox(int x1, int y1, int x2, int y2, T colour)
		{
			drawLine(x1,y1,x1,y2, colour);
			drawLine(x1,y1,x2,y1, colour);
			drawLine(x2,y1,x2,y2, colour);
			drawLine(x1,y2,x2,y2, colour);
		}
		
		void drawThickLine(int x1, int y1, int x2, int y2, T colour, int width)
		{
			int i,dx,dy,sdx,sdy,dxabs,dyabs,x,y,px,py;
			int p,q;

			dx=x2-x1;			/* the horizontal distance of the line */
			dy=y2-y1;			/* the vertical distance of the line */
			dxabs=abs(dx);
			dyabs=abs(dy);
			sdx=sgn(dx);
			sdy=sgn(dy);
			x=dyabs>>1;
			y=dxabs>>1;
			px=x1;
			py=y1;

			for (p=-width/2; p<width/2; p++)
				for (q=-width/2; q<width/2; q++)
					setPixel(px+p,py+q,colour);

			if (dxabs>=dyabs) /* the line is more horizontal than vertical */
			{				for(i=0;i<dxabs;i++)
				{
					y+=dyabs;
					if (y>=dxabs)
					{
						y-=dxabs;
						py+=sdy;
					}
					px+=sdx;

					for (p=-width/2; p<width/2; p++)
						for (q=-width/2; q<width/2; q++)
							setPixel(px+p,py+q,colour);
				}
			}
			else /* the line is more vertical than horizontal */
			{
				for(i=0;i<dyabs;i++)
				{
					x+=dxabs;
					if (x>=dyabs)
					{
						x-=dyabs;
						px+=sdx;
					}
					py+=sdy;
					
					for (p=-width/2; p<width/2; p++)
						for (q=-width/2; q<width/2; q++)
							setPixel(px+p,py+q,colour);

				}
			}
		}

	//operators--------------------------------------------------
	mm_image<T>& operator=( const mm_image<T> &_image )
	{
     // Only resize, if size has to change.
     if ( _image.getWidth() != getWidth() ||
          _image.getHeight() != getHeight() )
       data.resize( boost::extents[ _image.getHeight() ]
                                  [ _image.getWidth() ] );
     data = _image.data;
     return (*this);
	}

};



//!!!Those functions will need to be rewritten once image iterator would have been implemented
 
//usefull to put on one line a sequence of functions  
template< typename T, typename U >
mm_image< T > mm_convert_image( const mm_image< U > &image ) {
   mm_image< T > retVal;
   mm_convert_image< T, U >( retVal, image );
   return retVal;
}

//With this function it is possible not to allocate an image already allocated and not to
//specify the type. Is it faster? Does it matter in this case?  !!!benchmark needed?
template< typename T, typename U >
void mm_convert_image(mm_image<T> & dest ,const mm_image< U > &src ) {
   convert_trait<T,U> ct;
#ifndef NDEBUG
   ct.debug();
#endif
   dest.init( src.getWidth(), src.getHeight());
   const U *p = src.rawData().data();
   T *q = dest.rawData().data();
   for ( int i=0; i<(signed)src.rawData().num_elements(); i++ )
     *q++ = ct.convert(*p++);
}

}

#endif
