//
// Generic mm_image functions
//
// Bala Amavasai (bala@amavasai.org)
// Mon Oct  1 15:41:44 BST 2001
//
// $Header: /cvs/mimas/include/mm_image_funcs.h,v 1.26 2004/12/12 23:31:18 engjw Exp $
//

#ifndef MM_IMAGE_FUNCS_H
#define MM_IMAGE_FUNCS_H

#include <cassert>
#include <iostream>
#include <vector>
#include <algorithm>
#include <cstdlib>
#include <values.h>
#include <cmath>
#include "mm_property_image.h"
#include "mm_image.h"
#include "mm_rgba.h"
#include "mm_image_qtoutput.h"
#include "mm_prop.h"

namespace mimas {
void subtractImage(mm_image<int> &destimage, mm_image<int> &image1, mm_image<int> &image2);

template<typename T>
void normalise(mm_image<T> &imagein, T lower, T upper)
{
    T maxval=(T)MININT, minval=(T)MAXINT;

    for (int j=0; j<imagein.getHeight(); j++)
        for (int i=0; i<imagein.getWidth(); i++)
        {
            const T &value = imagein.pixel(i,j);
            if (value>maxval) maxval=value;
            if (value<minval) minval=value;
        }

    if (maxval==minval) return;

    //cerr << maxval << " " << minval << std::endl;

    for (int j=0; j<imagein.getHeight(); j++)
        for (int i=0; i<imagein.getWidth(); i++)
          imagein.pixel(i,j) =
            lower +
            (int)((float)(upper-lower)*(float)(imagein.pixel(i,j)-minval)/
                  (float)(maxval-minval));

}

template<typename T>
void normalise(mm_image<T> &imagein)
{
    normalise(imagein,(T)0,(T)255);
}


template<typename T>
void equalise(mm_image<T> &imagein)
{

    int H[256], TR[256];

    for (int count=0; count<256; count++)
	H[count]=0;

    int intensity;
    for (int j=0; j<imagein.getHeight(); j++)
        for (int i=0; i<imagein.getWidth(); i++)
	{
	    intensity=imagein.pixel(i,j);
       assert( intensity <= 255 );
	    H[intensity]+=1;
	}
    
   // convert to cumulative histogram 
    int imagesize=imagein.getWidth()*imagein.getHeight();
    for (int count=1; count<256; count++)
    {
	H[count]=H[count]+H[count-1];
	TR[count-1]=(255*H[count-1])/imagesize;
    }
    TR[255]=(255*H[255])/imagesize;
     
    for (int j=0; j<imagein.getHeight(); j++)
        for (int i=0; i<imagein.getWidth(); i++)
          imagein.pixel(i,j)=(T)TR[imagein.pixel(i,j)];

}



template<typename T>
void threshold(const mm_image<T> &imagein, mm_image<T> &imageout, T thresval, T outlevel1, T outlevel2)
{
  if ( &imageout != &imagein )
    imageout.init(imagein.getWidth(),imageout.getHeight());

    for (int y=0; y<imagein.getHeight(); y++)
        for (int x=0; x<imagein.getWidth(); x++)
            if (imagein.pixel(x,y)<thresval)
              imageout.pixel(x,y)=outlevel1;
            else
              imageout.pixel(x,y)=outlevel2;

}

template<typename T>
void threshold(mm_image<T> &imagein, T thresval)
{
    threshold(imagein,imagein,thresval,(T)0,(T)255);
}


template<typename T>
void threshold(mm_image<T> &imagein, mm_image<T> &imageout, T thresval)
{
    threshold(imagein,imageout,thresval,(T)0,(T)255);
}

template<typename T, typename U>
void despeckleKuwahara(const mm_image<T> &imagein, mm_image<U> &imageout)
{
  if ( &imageout != &imagein )
    imageout.init(imagein.getWidth(),imagein.getHeight());
    
    U intensity;
    float m_ul, m_ur, m_ll, m_lr;
    float v_ul, v_ur, v_ll, v_lr;
    
    for (int y=0; y<imagein.getHeight(); y++)
	for (int x=0; x<imagein.getWidth(); x++)
	{
	    if (y<2 || y>imagein.getHeight()-3 || x<2 || x>imagein.getWidth()-3)
	    {
         imageout.pixel(x,y)=imagein.pixel(x,y);
		continue;

	    }


	    // mean
	    m_ul=0.0; m_ur=0.0; m_ll=0.0; m_lr=0.0;

	    for (int j=0; j<3; j++)
		for (int i=0; i<3; i++)
		{
		    m_ul+=(float)imagein.pixel(x-2+i,y-2+j);
		    m_ur+=(float)imagein.pixel(x+i,y-2+j);
		    m_ll+=(float)imagein.pixel(x-2+i,y+j);
		    m_lr+=(float)imagein.pixel(x+i,y+j);
		}

	    m_ul=m_ul/9.0;
	    m_ur=m_ur/9.0;
	    m_ll=m_ll/9.0;
	    m_lr=m_lr/9.0;

	    
	    // estimated variance
	    v_ul=0.0; v_ur=0.0; v_ll=0.0; v_lr=0.0;
	    
	    for (int j=0; j<3; j++)
		for (int i=0; i<3; i++)
		{
		    v_ul+=(((float)imagein.pixel(x-2+i,y-2+j)-m_ul)*((float)imagein.pixel(x-2+i,y-2+j)-m_ul));
		    v_ur+=(((float)imagein.pixel(x+i,y-2+j)-m_ur)*((float)imagein.pixel(x+i,y-2+j)-m_ur));
		    v_ll+=(((float)imagein.pixel(x-2+i,y+j)-m_ll)*((float)imagein.pixel(x-2+i,y+j)-m_ll));
		    v_lr+=(((float)imagein.pixel(x+i,y+j)-m_lr)*((float)imagein.pixel(x+i,y+j)-m_lr));
		}

	    intensity=(U)m_ul;
	    intensity=v_ur<v_ul?(U)m_ur:intensity;
	    intensity=v_ll<v_ur?(U)m_ll:intensity;
	    intensity=v_lr<v_ll?(U)m_lr:intensity;

	    imageout.pixel(x,y)=intensity;

	}
  
}


template<typename T>
void despeckleKuwahara(mm_image<T> &imagein)
{
    mm_image<T> tempimage;

    despeckleKuwahara(imagein, tempimage);

    imagein = tempimage;
}

enum mm_filter_direction { _dir_x, _dir_y, _dir_xy};

template<typename T, typename U>
void gradSobel(mm_image<T> &imagein, mm_image<U> &imageout, mm_filter_direction dir)
{
    int	x,y, sx,sy;
    int	height, width;
    mm_image<U> tempimage;
    float	sxIntensity, syIntensity;


    static float sobelx[3][3]={	{ -1.0, -2.0,-1.0},
                                {  0.0,  0.0, 0.0},
                                {  1.0,  2.0, 1.0}};

    static float sobely[3][3]={ 	{-1.0, 0.0, 1.0},
                                 {-2.0, 0.0, 2.0},
                                 {-1.0, 0.0, 1.0}};

    height=imagein.getHeight();
    width=imagein.getWidth();
    tempimage.init(width,height);

    for(y=1;y<height-1;y++)
        for (x=1;x<width-1;x++)
        {
            sxIntensity=0.0; syIntensity=0.0;
            for(sx=0;sx<3;sx++)
                for(sy=0;sy<3;sy++)
                {
                    if ((dir==_dir_x) || (dir==_dir_xy)) sxIntensity+=sobelx[sy][sx]*(float)imagein.pixel(x+sx-1,y+sy-1);
                    if ((dir==_dir_y) || (dir==_dir_xy)) syIntensity+=sobely[sy][sx]*(float)imagein.pixel(x+sx-1,y+sy-1);
                }
            tempimage.pixel(x,y)=(U)sqrt((sxIntensity*sxIntensity)+(syIntensity*syIntensity));

        }

    imageout = tempimage;
}


template<typename T, typename U>
void edgeSobel(const mm_image<T> &imagein, mm_image<U> &imageout)
{
    int	x,y, sx,sy;
    int	height, width;
    mm_image<U> tempimage;
    float	sxIntensity, syIntensity;


    static float sobelx[3][3]={	{ -1.0, -2.0,-1.0},
                                {  0.0,  0.0, 0.0},
                                {  1.0,  2.0, 1.0}};

    static float sobely[3][3]={ 	{-1.0, 0.0, 1.0},
                                 {-2.0, 0.0, 2.0},
                                 {-1.0, 0.0, 1.0}};

    height=imagein.getHeight();
    width=imagein.getWidth();
    tempimage.init(width,height);

    for(y=1;y<height-1;y++)
        for (x=1;x<width-1;x++)
        {
            sxIntensity=0.0; syIntensity=0.0;
            for(sx=0;sx<3;sx++)
                for(sy=0;sy<3;sy++)
                {
                    sxIntensity+=sobelx[sy][sx]*(float)imagein.pixel(x+sx-1,y+sy-1);
                    syIntensity+=sobely[sy][sx]*(float)imagein.pixel(x+sx-1,y+sy-1);
                }
            tempimage.pixel(x,y)=(U)sqrt((sxIntensity*sxIntensity)+(syIntensity*syIntensity));

        }

    imageout = tempimage;
}

template<typename T>
void edgeSobel(mm_image<T> &imagein)
{
    mm_image<T> tempimage;

    edgeSobel(imagein,tempimage);

    imagein = tempimage;
}






template<typename T, typename U>
void edgeSobelOrientation(const mm_image<T> &imagein, mm_image<U> &imageout)
{
    int	x,y, sx,sy;
    int	height, width;
    mm_image<U> tempimage;
    float	sxIntensity, syIntensity;


    static float sobelx[3][3]={	{ -1.0, -2.0,-1.0},
                                {  0.0,  0.0, 0.0},
                                {  1.0,  2.0, 1.0}};

    static float sobely[3][3]={ 	{-1.0, 0.0, 1.0},
                                 {-2.0, 0.0, 2.0},
                                 {-1.0, 0.0, 1.0}};

    height=imagein.getHeight();
    width=imagein.getWidth();
    tempimage.init(width,height);

    for(y=1;y<height-1;y++)
        for (x=1;x<width-1;x++)
        {
            sxIntensity=0.0; syIntensity=0.0;
            for(sx=0;sx<3;sx++)
                for(sy=0;sy<3;sy++)
                {
                    sxIntensity+=sobelx[sy][sx]*(float)imagein.pixel(x+sx-1,y+sy-1);
                    syIntensity+=sobely[sy][sx]*(float)imagein.pixel(x+sx-1,y+sy-1);
                }
            if (sxIntensity==0.0)
              tempimage.pixel(x,y)=0.0;
            else
              tempimage.pixel(x,y)=atan(syIntensity/sxIntensity);


        }

    imageout = tempimage;
}

template<typename T, typename U>
void edgeLaplacian(const mm_image<T> &imagein, mm_image<U> &imageout)
{
    int	x,y, sx,sy;
    int	height, width;
    mm_image<U> tempimage;
    float	intensity;


    static float laplacian[3][3]={	{  0.0, -1.0, 0.0},
                                   { -1.0,  4.0,-1.0},
                                   {  0.0, -1.0, 0.0}};

    height=imagein.getHeight();
    width=imagein.getWidth();
    tempimage.init(width,height);

    for(y=1;y<height-1;y++)
        for (x=1;x<width-1;x++)
        {
            intensity=0.0;
            for(sx=0;sx<3;sx++)
                for(sy=0;sy<3;sy++)
                    intensity+=laplacian[sy][sx]*(float)imagein.pixel(x+sx-1,y+sy-1);
            tempimage.pixel(x,y)=(U)fabs(intensity);

        }

    imageout = tempimage;
}

template<typename T>
void edgeLaplacian(mm_image<T> &imagein)
{
    mm_image<T> tempimage;

    edgeLaplacian(imagein,tempimage);

    imagein = tempimage;
}

template<typename T, typename U>
void edgeLoG(const mm_image<T> &imagein, mm_image<U> &imageout)
{
    int	x,y, sx,sy;
    int	height, width;
    mm_image<float> tempimage;
    float	intensity;


    static float LoG[9][9]={{   0,  0,  3,  2,  2,  2,  3,  0,  0},
                            {   0,  2,  3,  5,  5,  5,  3,  2,  0},
                            {   3,  3,  5,  3,  0,  3,  5,  3,  3},
                            {   2,  5,  3,-12,-23,-12,  3,  5,  2},
                            {   2,  5,  0,-23,-40,-23,  0,  5,  2},
                            {   2,  5,  3,-12,-23,-12,  3,  5,  2},
                            {   3,  3,  5,  3,  0,  3,  5,  3,  3},
                            {   0,  2,  3,  5,  5,  5,  3,  2,  0},
                            {   0,  0,  3,  2,  2,  2,  3,  0,  0}};

    /*static float LoG[5][5]={{  0.0, 0.0,-1.0, 0.0, 0.0},
                            {  0.0,-1.0,-2.0,-1.0, 0.0},
                            { -1.0,-2.0,16.0,-2.0,-1.0},
                            {  0.0,-1.0,-2.0,-1.0, 0.0},
                            {  0.0, 0.0,-1.0, 0.0, 0.0}};
    */
    height=imagein.getHeight();
    width=imagein.getWidth();
    tempimage.init(width,height);

    for(y=4;y<height-4;y++)
        for (x=4;x<width-4;x++)
        {
            intensity=0.0;
            for(sx=0;sx<9;sx++)
                for(sy=0;sy<9;sy++)
                    intensity-=LoG[sy][sx]*(float)imagein.pixel(x+sx-4,y+sy-4);
            tempimage.pixel(x,y)=(U)(intensity);

        }

    threshold(tempimage,(float)0.0);

    imageout.init( width, height );

    // Shouldn't actually threshold - must check zero crossings
    for (y=4; y<height-4; y++)
        for (x=4; x<width-4; x++)
        {
            if (tempimage.pixel(x,y)<=127) continue;

            if ((tempimage.pixel(x-1,y-1)>127) ||
                    (tempimage.pixel(x-1,y)>127) ||
                    (tempimage.pixel(x-1,y+1)>127) ||
                    (tempimage.pixel(x,y-1)>127) ||
                    (tempimage.pixel(x,y+1)>127) ||
                    (tempimage.pixel(x+1,y-1)>127) ||
                    (tempimage.pixel(x+1,y)>127) ||
                    (tempimage.pixel(x+1,y+1)>127))
              imageout.pixel(x,y)=(U)255;
        }
}

template<typename T>
void edgeLoG(mm_image<T> &imagein)
{
    mm_image<T> tempimage;

    edgeLoG(imagein, tempimage);

    imagein = tempimage;
}



// IMPORTANT NOTE
// imagein1 is image at time t-1
// imagein2 is image at time t
template<typename T, typename U>
void edgeHaynesJain(const mm_image<T> &imagein1, const mm_image<T> &imagein2,
                    mm_image<U> &imageout)
{
    int width, height;

    width=imagein1.getWidth();
    height=imagein1.getHeight();

    MMERROR( imagein2.getWidth() == width && imagein2.getHeight() ==height,
             mm_exception, , "mm_image_funcs::edgeHaynesJain - "
             "size of input images do not match" );

    if ( &imageout != &imagein1 && &imageout != &imagein2 )
      imageout.init( width, height );

    //mm_image<T> edge1, edge2;
    mm_image<T> edge2;
    //edgeSobel(imagein1, edge1);
    edgeSobel(imagein2, edge2);

    for (int y=1; y<height-1; y++)
        for (int x=1; x<width-1; x++)
          imageout.pixel(x,y)=(U)fabs((double)edge2.pixel(x,y)*(double)(imagein2.pixel(x,y)-imagein1.pixel(x,y)));

}


template<typename T, typename U>
void halfResolution(const mm_image<T> &imagein, mm_image<U> &imageout)
{
    // note that this is only an estimate!
    static float gauss[3][3]={	{  1.0,  2.0, 1.0},
                               {  2.0,  4.0, 2.0},
                               {  1.0,  2.0, 1.0}};
    static float gaussfac=(1.0/16.0);

    assert( &imageout != &imagein );
    imageout.init(imagein.getWidth()/2, imagein.getHeight()/2);

    for (int j=1; j<imagein.getHeight(); j+=2)
        for (int i=1; i<imagein.getWidth(); i+=2)
        {
            float tempval=0.0;
            for (int y=-1; y<=1; y++)
                for (int x=-1; x<=1; x++)
                    tempval+=(imagein.pixel(i+x,j+y)*gauss[y+1][x+1]);
            imageout.pixel(i/2,j/2)=(U)(tempval*gaussfac);
        }
}


template<typename T>
void halfResolution(mm_image<T> &imagein)
{
    mm_image<T> tempimage;

    halfResolution(imagein, tempimage);

    imagein = tempimage;
}

// Note this assumes black edges on white background unlike halfResolution
// above
template<typename T, typename U>
void halfResolutionEdgePyramid(const mm_image<T> &imagein, mm_image<U> &imageout)
{
  assert( &imageout != &imagein );
    imageout.init(imagein.getWidth()/2, imagein.getHeight()/2);

    for (int j=1; j<imagein.getHeight(); j+=2)
        for (int i=1; i<imagein.getWidth(); i+=2)
        {
            int tempval=0;
	    for (int y=-1; y<=1; y++)
                for (int x=-1; x<=1; x++)
                    if (imagein.getPixel(i+x,j+y)>127)
			tempval=255;
	    
	    imageout.pixel(i/2,j/2)=(U)tempval;
        }
}


template<typename T>
void halfResolutionEdgePyramid(mm_image<T> &imagein)
{
    mm_image<T> tempimage;

    halfResolutionEdgePyramid(imagein, tempimage);

    imagein = tempimage;
}

// 3x3 median despecking filter
template<typename T>
void despeckleMedian(const mm_image<T> &imagein, mm_image<T> &imageout)
{
  std::vector<T> pixvals;
  assert( &imageout != &imagein );
  imageout.init(imagein.getWidth(),imagein.getHeight());
 
  for (int j=1; j<imagein.getHeight()-1; j++)
	for (int i=1; i<imagein.getWidth()-1; i++)
	{
	    if (!pixvals.empty()) pixvals.clear();
	    for (int y=-1; y<=1; y++)
		for (int x=-1; x<=1; x++)
		    pixvals.push_back(imagein.pixel(i+x,j+y));
	    
	    sort(pixvals.begin(), pixvals.end());
	    imageout.pixel(i,j)=pixvals[pixvals.size()/2];
	}
 
    for (int j=0; j<imagein.getHeight(); j++)
    {
      imageout.pixel(0,j) = imagein.pixel(0,j);
      imageout.pixel(imagein.getWidth()-1,j)=imagein.pixel(imagein.getWidth()-1,j);
    }
    
    for (int i=0; i<imagein.getWidth(); i++)
    {
    	imageout.pixel(i,0)=imagein.pixel(i,0);
      imageout.pixel(i,imagein.getHeight()-1)=imagein.pixel(i,imagein.getHeight()-1);
    }
}
 
template<typename T>
void despeckleMedian(mm_image<T> &imagein)
{
    mm_image<T> tempimage;

    despeckleMedian(imagein, tempimage);

    imagein = tempimage;
}

void invert(mm_image<int> &inimage);

template<typename T>
void chamfer(mm_image<T> &inimage)
{
    int nw,n,ne,w,e,sw,s,se;
    int fp;

    for (int j=1; j<inimage.getHeight()-1; j++)
        for (int i=1; i<inimage.getWidth(); i++)
        {
            fp=inimage.pixel(i,j);
            nw=4+inimage.pixel(i-1,j-1);
            n=3+inimage.pixel(i,j-1);
            w=3+inimage.pixel(i-1,j);
            sw=4+inimage.pixel(i-1,j+1);

            fp=fp<nw?fp:nw;
            fp=fp<n?fp:n;
            fp=fp<w?fp:w;
            fp=fp<sw?fp:sw;
            inimage.pixel(i,j)=fp;

        }

    for (int j=inimage.getHeight()-2; j>0; j--)
        for (int i=inimage.getWidth()-2; i>=0; i--)
        {
            fp=inimage.pixel(i,j);
            ne=4+inimage.pixel(i+1,j-1);
            e=3+inimage.pixel(i+1,j);
            s=3+inimage.pixel(i,j+1);
            se=4+inimage.pixel(i+1,j+1);

            fp=fp<ne?fp:ne;
            fp=fp<e?fp:e;
            fp=fp<s?fp:s;
            fp=fp<se?fp:se;
            inimage.pixel(i,j)=fp;

        }
    
    for (int i=0; i<inimage.getWidth(); i++)
    {
      inimage.pixel(i,0)=inimage.pixel(i,1);
      inimage.pixel(i,inimage.getHeight()-1)=inimage.pixel(i,inimage.getHeight()-2);
    }
    
    for (int j=0; j<inimage.getHeight(); j++)
    {
      inimage.pixel(0,j)=inimage.pixel(1,j);
      inimage.pixel(inimage.getWidth()-1,j)=inimage.pixel(inimage.getWidth()-2,j);
    }
 }


template<typename T>
void paethRotateRadian(const mm_image<T> &imagein, mm_image<T> &imageout,
                       float radian)
{
  assert( &imageout != &imagein );
    imageout.init(imagein.getWidth(),imagein.getHeight());

    if (fabs(radian-M_PI)<1e-3) radian=radian<M_PI?M_PI-1e-3:M_PI+1e-3; // Quick hack to fix the problem when tan goes to infinity
    float s1=-tan(-radian/2.0);
    float s2=sin(-radian);

    float xr, yr;
    int x,y;
    for (int j=0; j<imageout.getHeight(); j++)
        for (int i=0; i<imageout.getWidth(); i++)
        {
            x=i-imageout.getWidth()/2;
            y=j-imageout.getHeight()/2;
            //x=i;
            //y=j;

            xr = (float)x + s1 * (float)y;
            yr = (float)y + s2 * xr;
            xr = xr + s1 * yr;

            xr+=(float)imageout.getWidth()/2;
            yr+=(float)imageout.getHeight()/2;
            imageout.pixel(i, j)=imagein.getPixel((int)xr,(int)yr);
            //imageout.setPixel(i,j, imagein.getPixel(i,j));
        }
}

template<typename T>
void multiLevelCorrelation(mm_image<T> &edgemap, mm_image<T> &objectedgemap, int
                           levels, int &x, int &y, float &corrval, bool
			   chamferObject)
{
    mm_image<T> object[levels], image[levels];
    
    // set background colour of object to be white
    for (int l=0; l<levels; l++)
    {
	object[l].setInitVal(255);
	image[l].setInitVal(255);
    }

    image[0] = edgemap;
   // invert(image[0]);

    object[0] = objectedgemap;
   // invert(object[0]);
  
     for (int l=1; l<levels; l++)
    {
        halfResolutionEdgePyramid(image[l-1],image[l]);
        halfResolutionEdgePyramid(object[l-1],object[l]);
    }
    
    for (int l=0; l<levels; l++)
    {
	invert(image[l]);
	invert(object[l]);
	//image[l].display();
	//object[l].display();
	chamfer(image[l]);
	if (chamferObject) chamfer(object[l]);
    }
 
     int xmin = INT_MIN, ymin = INT_MIN;
    int xminsearch, xmaxsearch, yminsearch, ymaxsearch;
    float corr, mincorr;

    x=0; y=0;
    xminsearch=0;
    yminsearch=0;
    xmaxsearch=image[levels-1].getWidth();
    ymaxsearch=image[levels-1].getHeight();

    for (int l=levels-1; l>=0; l--)
    {
	//object[l].display();
	//image[l].display();

        mincorr=FLT_MAX;
        for (int j=yminsearch; j<ymaxsearch-object[l].getHeight(); j++)
            for (int i=xminsearch; i<xmaxsearch-object[l].getWidth(); i++)
            {
                corr=0.0;
		int count=0;
                for (int b=0; b<object[l].getHeight(); b++)
                    for (int a=0; a<object[l].getWidth(); a++)
                    {
                        if (!chamferObject) if (object[l].pixel(a,b)>127) continue;
                        corr+=(float)abs((int)image[l].pixel(i+a,j+b)-(int)object[l].pixel(a,b));
			count++;
                    }

		corr/=count;
                if (corr<mincorr)
                {
                    xmin=i; ymin=j;
                    mincorr=corr;
                }
            }

        if (l!=0)
        {
            xminsearch=(xmin*2)-1-20;
            yminsearch=(ymin*2)-1-20;

            xmaxsearch=(xmin*2)-1+object[l-1].getWidth()+20;
            ymaxsearch=(ymin*2)-1+object[l-1].getHeight()+20;

            if (xminsearch<0) xminsearch=0;
            if (yminsearch<0) yminsearch=0;

            if (xmaxsearch>image[l-1].getWidth()) xmaxsearch=image[l-1].getWidth();
            if (ymaxsearch>image[l-1].getHeight()) ymaxsearch=image[l-1].getHeight();
        }


    }

    assert( xmin != INT_MIN && ymin != INT_MIN );
    x=xmin;
    y=ymin;
    corrval=mincorr;

}

template<typename T>
void correlation(const mm_image<T> &imagein, mm_image<T> &templatein,
                 int &x, int &y, float &corrval)
{

    int xmin = INT_MIN, ymin = INT_MIN;
    int xminsearch, xmaxsearch, yminsearch, ymaxsearch;
    float corr, mincorr;

    x=0; y=0;
    xminsearch=0;
    yminsearch=0;
    xmaxsearch=imagein.getWidth();
    ymaxsearch=imagein.getHeight();

    mincorr=FLT_MAX;
    for (int j=yminsearch; j<ymaxsearch-templatein.getHeight(); j++)
        for (int i=xminsearch; i<xmaxsearch-templatein.getWidth(); i++)
        {
            corr=0.0;
            for (int b=0; b<templatein.getHeight(); b++)
                for (int a=0; a<templatein.getWidth(); a++)
                {
                    //if (object[l].getPixel(a,b)>127) continue;
                    corr+=(float)abs((int)imagein.pixel(i+a,j+b)-(int)templatein.pixel(a,b));
                }

            if (corr<mincorr)
            {
                xmin=i; ymin=j;
                mincorr=corr;
            }
        }

    assert( xmin != INT_MIN && ymin != INT_MIN );
    x=xmin;
    y=ymin;
    corrval=mincorr;

}

template<typename T>
void correlationChamfer(const mm_image<T> &imagein, mm_image<T> &templatein,
                        int &x, int &y, float &corrval)
{

    int xmin = INT_MIN, ymin = INT_MIN;
    int xminsearch, xmaxsearch, yminsearch, ymaxsearch;
    float corr, mincorr;

    x=0; y=0;
    xminsearch=0;
    yminsearch=0;
    xmaxsearch=imagein.getWidth();
    ymaxsearch=imagein.getHeight();

    mincorr=FLT_MAX;
    for (int j=yminsearch; j<ymaxsearch-templatein.getHeight(); j++)
        for (int i=xminsearch; i<xmaxsearch-templatein.getWidth(); i++)
        {
            corr=0.0;
            for (int b=0; b<templatein.getHeight(); b++)
                for (int a=0; a<templatein.getWidth(); a++)
                {
		    if (templatein.pixel(a,b)>127) continue;
                    corr+=(float)abs((int)imagein.pixel(i+a,j+b)-(int)templatein.pixel(a,b));
                }

            if (corr<mincorr)
            {
                xmin=i; ymin=j;
                mincorr=corr;
            }
        }

    assert( xmin != INT_MIN && ymin != INT_MIN );
    x=xmin;
    y=ymin;
    corrval=mincorr;

}

template<typename T>
void rotationChamferCorrelationEdgeMap(mm_image<T> &edgemap, mm_image<T>
	&objectedgemap, int angle1, int angle2, int anglestep, int &x, int &y, int &outangle, float &corrval)
{
    mm_image<T> chamferedImage(255), rotatedObject(255), chamferedObject(255);

    chamferedImage = edgemap;
    invert(chamferedImage);
    chamfer(chamferedImage);

  
    invert(objectedgemap);
    // Make sure that object boundaries are white:
    objectedgemap.drawLine(0,0,0,objectedgemap.getHeight()-1,255);
    objectedgemap.drawLine(objectedgemap.getWidth()-1,0,objectedgemap.getWidth()-1,objectedgemap.getHeight()-1,255);
    objectedgemap.drawLine(0,0,objectedgemap.getWidth()-1,objectedgemap.getHeight()-1,255);
    objectedgemap.drawLine(0,objectedgemap.getHeight()-1,objectedgemap.getWidth()-1,objectedgemap.getHeight()-1,255);
    //chamfer(objectedgemap);

    float cval;
    int tx,ty;
    corrval=FLT_MAX;
    for (int angle=angle1; angle<=angle2; angle+=anglestep)
    {
   rotatedObject = objectedgemap;
	//invert(rotatedObject);
	paethRotateRadian(rotatedObject,chamferedObject,(float)angle*M_PI/180.0);
	//chamfer(chamferedObject);
	correlationChamfer(chamferedImage, chamferedObject, tx, ty, cval);

	if (cval<corrval)
	{
	    corrval=cval;
	    x=tx;
	    y=ty;
	    outangle=angle;
	}
    }
}


template<typename T>
  void convolve(const mm_image<T> &imagein, mm_image<T> &imageout,
                std::vector < std::vector<float> > &filter, float divcoef,
                float bias)
{
    const int fy=filter.size();
    const int fx=filter[0].size();
    
    assert( &imageout != &imagein );
    imageout.init(imagein.getWidth(),imagein.getHeight());
    
    float intensity;
    for (int y=0; y<imagein.getHeight(); y++)
	for (int x=0; x<imagein.getWidth(); x++)
	{
	    if (y-fy/2<0 || y+fy/2>imagein.getHeight()-1 || x-fx/2<0 || x+fx/2>imagein.getWidth()-1)
	    {
         imageout.pixel(x,y)=imagein.pixel(x,y);
		continue;
	    }

	    intensity=0.0;
	    for (int j=-fy/2; j<=fy/2; j++)
		for (int i=-fx/2; i<=fx/2; i++)
		    intensity+=filter[j+fy/2][i+fx/2]*imagein.pixel(x+i,y+j);
	    imageout.pixel(x,y)=(T)(bias+intensity/divcoef);
	}

    
}

template<typename T>
void focusEnhance(const mm_image<T> &imagein, mm_image<T> &imageout)
{
  std::vector < std::vector<float> > f(3,3);

    f[0][0]=-1;  f[0][1]=0;  f[0][2]=-1;
    f[1][0]= 0;  f[1][1]=7;  f[1][2]=0;
    f[2][0]=-1;  f[2][1]=0;  f[2][2]=-1;
   
    convolve(imagein, imageout, f, 3.0, 0.0);
    normalise(imageout);
}


template<typename T>
void featureEnhance(const mm_image<T> &imagein, mm_image<T> &imageout)
{
  std::vector < std::vector<float> > f(3,3);

    f[0][0]= 0;  f[0][1]=-1;  f[0][2]= 0;
    f[1][0]=-1;  f[1][1]=10;  f[1][2]=-1;
    f[2][0]= 0;  f[2][1]=-1;  f[2][2]= 0;
   
    convolve(imagein, imageout, f, 6.0, 0.0);
    normalise(imageout);
}


template<typename T>
void emboss(const mm_image<T> &imagein, mm_image<T> &imageout)
{
  std::vector < std::vector<float> > f(3,3);

    f[0][0]=-1;  f[0][1]=0;  f[0][2]=0;
    f[1][0]= 0;  f[1][1]=0;  f[1][2]=0;
    f[2][0]= 0;  f[2][1]=0;  f[2][2]=1;
   
    convolve(imagein, imageout, f, 1.0, 0.0);
    normalise(imageout);
}

template<typename T>
void softenHeavy(const mm_image<T> &imagein, mm_image<T> &imageout)
{
  std::vector < std::vector<float> > f(3,3);

    f[0][0]=11;  f[0][1]=11;  f[0][2]=11;
    f[1][0]=11;  f[1][1]=11;  f[1][2]=11;
    f[2][0]=11;  f[2][1]=11;  f[2][2]=11;
   
    convolve(imagein, imageout, f, 99.0, 0.0);
    normalise(imageout);
}

template<typename T>
void softenMedium(const mm_image<T> &imagein, mm_image<T> &imageout)
{
  std::vector < std::vector<float> > f(3,3);

    f[0][0]=10;  f[0][1]=10;  f[0][2]=10;
    f[1][0]=10;  f[1][1]=20;  f[1][2]=10;
    f[2][0]=10;  f[2][1]=10;  f[2][2]=10;
   
    convolve(imagein, imageout, f, 100.0, 0.0);
    normalise(imageout);
}

template<typename T>
void softenLight(const mm_image<T> &imagein, mm_image<T> &imageout)
{
  std::vector < std::vector<float> > f(3,3);

    f[0][0]= 6;  f[0][1]=12;  f[0][2]= 6;
    f[1][0]=12;  f[1][1]=25;  f[1][2]=12;
    f[2][0]= 6;  f[2][1]=12;  f[2][2]= 6;
   
    convolve(imagein, imageout, f, 97.0, 0.0);
    normalise(imageout);
}

template<typename T>
void oilPainting(const mm_image<T> &imagein, mm_image<T> &imageout,
                 int regiondim)
{
   const int fy=regiondim;
    const int fx=regiondim;

    assert( &imageout != &imagein );
    imageout.init(imagein.getWidth(),imagein.getHeight());
    
    for (int y=0; y<imagein.getHeight(); y++)
	for (int x=0; x<imagein.getWidth(); x++)
	{
	    if (y-fy/2<0 || y+fy/2>imagein.getHeight()-1 || x-fx/2<0 || x+fx/2>imagein.getWidth()-1)
	    {
         imageout.pixel(x,y)=imagein.pixel(x,y);
		continue;
	    }

       std::vector<int> pixcolor(regiondim*regiondim);
       std::vector<int> graylevel(256);
    
	    int count=0;
	    for (int j=-fy/2; j<=fy/2; j++)
		for (int i=-fx/2; i<=fx/2; i++)
		{
		    pixcolor[count]=imagein.pixel(x+i,y+j);
		    graylevel[pixcolor[count]]+=1;
		    
		    count++;
		}
	    
	    count=0;
	    int intensity=0, maxuse=0;
	    for (int j=-fy/2; j<=fy/2; j++)
		for (int i=-fx/2; i<=fx/2; i++)
		{
		    if (graylevel[pixcolor[count]]>maxuse)
		    {
			maxuse=graylevel[pixcolor[count]];
			intensity=pixcolor[count];
		    }

		    count++;
		}
	    
	    imageout.pixel(x,y)=(T)(intensity);
	}

    
}

template<typename T>
void displayWithOverlay( mm_image_output< mm_rgba< unsigned char > > &display,
                         mm_image<T>& im, mm_property_image& pim)
{
    //alas can't be part of the base class. circular dependencies :-(
    mm_image< mm_rgba< unsigned char > > output;
    output.init( im.getWidth(), im.getHeight() );
    
    for( int j=0;j<im.getHeight();j++)
      for( int i=0;i<im.getWidth();i++) {
        mm_prop_ptr p( pim.pixel(i,j) );
        if( p )
          p->draw( output, i, j, false);
      }
    
    for( int j=0;j<im.getHeight();j++)
      for( int i=0;i<im.getWidth();i++)
        if ( (int)output.pixel( i, j ) ==  0 ) {
          unsigned char value = (unsigned char)im.pixel( i, j );
          output.pixel( i, j ) =
            mm_rgba< unsigned char >( value, value, value );
        };
          
    display << output;
}

}

#endif
