#ifndef MM_IMAGE_OUTPUT_H
#define MM_IMAGE_OUTPUT_H

#include "mm_image.h"

namespace mimas {

/** @addtogroup imageIO
    @{ */
/** @defgroup imageOutput Image Output
    This group contains classes for writing image(s) to different output sinks.
    @{ */
/** Abstract base-class for output sinks.
    This is a base-class for output sinks. It allows to implement a program,
    which can write image(s) to an arbitrary output-sink. F.e.:
    \code
    using namespace mimas;
    ...
    void test( mm_image_output< mm_rgba< unsigned char > > &sink ) {
       mm_image< mm_rgba< unsigned char > > image;
       ...
       sink << image;
       ...
    };
    ...
    \endcode */
template< typename T >
class mm_image_output: public mm_object
{
 public:
  /// Constructor.
  mm_image_output(void) {}
  /** Write image to arbitrary sink.
      This function is virtual and has to be overriden by the inheriting class.
      @param image The image to write to the sink. */
  virtual void write( const mm_image< T > &image ) = 0;
};

/// Stream-operator for arbitrary output-sink.
template< typename T >
inline mm_image_output< T > &operator<<( mm_image_output< T > &output,
                                         const mm_image< T > &image )
{ output.write( image ); return output; }

///@}

///@}

};

#endif
