#ifndef __MM_MULTI_ARRAY_OP_H
#define __MM_MULTI_ARRAY_OP_H

#include <boost/mpl/apply_if.hpp>
#include <boost/mpl/equal_to.hpp>
#include <boost/multi_array.hpp>
#include <functional>

// scalar = f( scalar )
template< class _F >
struct mm_element_apply_intern:
  public std::unary_function< typename _F::argument_type,
                              typename _F::result_type >
{
  typedef mm_element_apply_intern< _F > type;
  typename _F::result_type& operator()
    ( typename _F::argument_type & __x ) {
    __x = f( __x );
    return __x;
  }
  _F f;
  };

// scalar1 = f( scalar1, scalar2 )
template< class _F >
struct mm_element_scalar_apply_intern:
  public std::binary_function< typename _F::first_argument_type,
                               typename _F::second_argument_type,
                               typename _F::result_type >
{
  typedef mm_element_scalar_apply_intern< _F > type;
  typename _F::result_type& operator()
    ( typename _F::first_argument_type & __x,
      const typename _F::second_argument_type & __s ) {
    __x = f( __x, __s );
    return __x;
  }
  _F f;
};

// array = f( array, scalar )
template < class _Array, class _F >
struct mm_array_scalar_apply_intern:
  public std::binary_function< _Array, typename _F::second_argument_type,
                               _Array >
{
  typedef mm_array_scalar_apply_intern< _Array, _F > type;
  _Array& operator()( _Array & __x,
                      const typename _F::second_argument_type& __s ) {
    for ( typename _Array::iterator i = __x.begin(); i != __x.end(); i++ ) {
      typename _Array::reference r = *i;
      f( r, __s );
    };
    return __x;
  }
  typename boost::mpl::apply_if
    < boost::is_same
      < typename _Array::value_type, typename _Array::element >,
        mm_element_scalar_apply_intern< _F >,
        mm_array_scalar_apply_intern< typename _Array::reference, _F >
      >::type f;
};

// array1 = f( array1, array2 )
template < class _Array1, class _Array2, class _F >
struct mm_array_array_apply_intern:
  public std::binary_function< _Array1, _Array2, _Array1 >
{
  typedef mm_array_array_apply_intern< _Array1, _Array2, _F > type;
  _Array1& operator()( _Array1 & __x, const _Array2& __y ) {
    assert( __x.num_dimensions() == __y.num_dimensions() );
    assert( std::equal( __x.shape(), __x.shape() + __x.num_dimensions(),
                        __y.shape() ) );
    typename _Array2::const_iterator j = __y.begin();
    for ( typename _Array1::iterator i = __x.begin(); i != __x.end();
          i++, j++ ) {
      typename _Array1::reference r = *i;
      f( r, *j );
    };
    return __x;
  }
  typename boost::mpl::apply_if
    < boost::is_same
      < typename _Array1::value_type, typename _Array1::element >,
        mm_element_scalar_apply_intern< _F >,
        mm_array_array_apply_intern< typename _Array1::reference,
                                     typename _Array2::const_reference, _F >
      >::type f;
};

// array = f( array )
template < class _Array, class _F >
struct mm_array_apply_intern:
  public std::unary_function< _Array, _Array >
{
  typedef mm_array_apply_intern< _Array, _F > type;
  _Array& operator()( _Array & __x ) {
    for ( typename _Array::iterator i = __x.begin(); i != __x.end(); i++ ) {
      typename _Array::reference r = *i;
      f( r );
    };
    return __x;
  }
  typename boost::mpl::apply_if
    < boost::is_same
      < typename _Array::value_type, typename _Array::element >,
        mm_element_apply_intern< _F >,
        mm_array_apply_intern< typename _Array::reference, _F >
      >::type f;
      };


// f( array )
template < class _Array, class _F >
struct mm_array_apply_extern:
  public std::unary_function
    < _Array, boost::multi_array< typename _F::result_type,
                                  _Array::dimensionality > >
{
  typedef mm_array_apply_extern
    < _Array, boost::multi_array< typename _F::result_type,
                                  _Array::dimensionality > > type;
  boost::multi_array< typename _F::result_type, _Array::dimensionality >
  operator()( const _Array & __x ) {
    boost::array< std::size_t, _Array::dimensionality > shape;
    boost::copy_n( __x.shape(), _Array::dimensionality, shape.begin() );
    boost::multi_array< typename _F::result_type, _Array::dimensionality >
      retval( shape );
    return f( retval, __x );
  };
  struct mm_ignore_first: public
     std::binary_function< typename _F::result_type,
                           typename _Array::element,
                           typename _F::result_type > {
    typename _F::result_type &operator()( typename _F::result_type &__x,
                                          const typename _Array::element &__c )
    {
      __x = f( __c );
      return __x;
    };
    _F f;
  };
  mm_array_array_apply_intern< boost::multi_array< typename _F::result_type,
                                                   _Array::dimensionality >,
    _Array, mm_ignore_first > f;
};

// f( array, scalar )
template < class _Array, class _F >
struct mm_array_scalar_apply_extern:
  public std::binary_function
  < _Array, typename _F::second_argument_type,
    boost::multi_array< typename _Array::element,
                        _Array::dimensionality > >
{
  boost::multi_array< typename _Array::element,
                      _Array::dimensionality >
  operator()( const _Array & __x,
              const typename _F::second_argument_type& __s ) {
    boost::array< std::size_t, _Array::dimensionality > shape;
    boost::copy_n( __x.shape(), _Array::dimensionality, shape.begin() );
    boost::multi_array< typename _Array::element, _Array::dimensionality >
      retval( shape );
    retval = __x;
    return f( retval, __s );
  }
  mm_array_scalar_apply_intern< boost::multi_array< typename _Array::element,
                                                    _Array::dimensionality >,
                                _F > f;
};

// f( array1, array2 )
template < class _Array1, class _Array2, class _F >
struct mm_array_array_apply_extern:
  public std::binary_function< _Array1, _Array2,
                               boost::multi_array< typename _Array1::element,
                                                   _Array1::dimensionality > >
{
  boost::multi_array< typename _Array1::element, _Array1::dimensionality >
  operator()( const _Array1 & __x, const _Array2& __y ) {
    boost::array< std::size_t, _Array1::dimensionality > shape;
    boost::copy_n( __x.shape(), _Array1::dimensionality, shape.begin() );
    boost::multi_array< typename _Array1::element, _Array1::dimensionality >
      retval( shape );
    retval = __x;
    f( retval, __y );
    return retval;
  }
  mm_array_array_apply_intern< boost::multi_array< typename _Array1::element,
                                                   _Array1::dimensionality >,
                               _Array2, _F > f;
};

// array = f( array, scalar )
#define MM_ARRAY_SCALAR_OP_INTERN( ARRAYTYPE, MODIFIER, NAME,         \
                                   OBJECT )                           \
  template< typename T, std::size_t N >                               \
  ARRAYTYPE< T, N > MODIFIER NAME ( ARRAYTYPE< T, N > MODIFIER x,     \
                                    const T &c ) {                    \
    mm_array_scalar_apply_intern< ARRAYTYPE< T, N >, OBJECT< T > > f; \
    return f( x, c );                                                 \
  }

// f( array, scalar )
#define MM_ARRAY_SCALAR_OP_EXTERN_LEFT( ARRAYTYPE, MODIFIER, NAME,     \
                                        OBJECT )                       \
  template< typename T, std::size_t N >                                \
  boost::multi_array< T, N > NAME( const ARRAYTYPE< T, N > MODIFIER x, \
                                   const T &c ) {                      \
    mm_array_scalar_apply_extern< ARRAYTYPE< T, N >, OBJECT< T > > f;  \
    return f( x, c );                                                  \
  }

// f( scalar, array )
#define MM_ARRAY_SCALAR_OP_EXTERN_RIGHT( ARRAYTYPE, MODIFIER, NAME,       \
                                         OBJECT )                         \
  template< typename T, std::size_t N >                                   \
  boost::multi_array< T, N > NAME( const T &c,                            \
                                   const ARRAYTYPE< T, N > MODIFIER x ) { \
    mm_array_scalar_apply_extern< ARRAYTYPE< T, N >, OBJECT< T > > f;     \
    return f( x, c );                                                     \
  }

// array = f( array, scalar ); f( array, scalar )
#define MM_ARRAY_SCALAR_OP_NONASSOC_NONCONST( ARRAYTYPE, MODIFIER,      \
                                              INAME, ENAME, OBJECT )    \
  MM_ARRAY_SCALAR_OP_INTERN     ( ARRAYTYPE, MODIFIER, INAME, OBJECT )  \
  MM_ARRAY_SCALAR_OP_EXTERN_LEFT( ARRAYTYPE, MODIFIER, ENAME, OBJECT )

// array = f( array, scalar ); f( array, scalar ); f( scalar, array )
#define MM_ARRAY_SCALAR_OP_ASSOC_NONCONST( ARRAYTYPE, MODIFIER,      \
                                           INAME, ENAME, OBJECT )    \
  MM_ARRAY_SCALAR_OP_NONASSOC_NONCONST( ARRAYTYPE, MODIFIER,         \
                                        INAME, ENAME, OBJECT )       \
  MM_ARRAY_SCALAR_OP_EXTERN_RIGHT     ( ARRAYTYPE, MODIFIER,         \
                                        ENAME, OBJECT )

// f( array, scalar )
#define MM_ARRAY_SCALAR_OP_NONASSOC_CONST( ARRAYTYPE, MODIFIER,        \
                                           INAME, ENAME, OBJECT )      \
  MM_ARRAY_SCALAR_OP_EXTERN_LEFT( ARRAYTYPE, MODIFIER, ENAME, OBJECT )

// f( array, scalar ); f( scalar, array )
#define MM_ARRAY_SCALAR_OP_ASSOC_CONST( ARRAYTYPE, MODIFIER,            \
                                        INAME, ENAME, OBJECT )          \
  MM_ARRAY_SCALAR_OP_NONASSOC_CONST( ARRAYTYPE, MODIFIER, INAME, ENAME, \
                                     OBJECT )                           \
  MM_ARRAY_SCALAR_OP_EXTERN_RIGHT  ( ARRAYTYPE, MODIFIER, ENAME, OBJECT )

// Branch depending on type of operator ( NONASSOC or ASSOC )
#define MM_ARRAY_SCALAR_OP( INAME, ENAME, OBJECT, TYPE )                  \
  MM_ARRAY_SCALAR_OP_##TYPE##_NONCONST                                    \
  ( boost::multi_array                                , &, INAME, ENAME,  \
    OBJECT )                                                              \
  MM_ARRAY_SCALAR_OP_##TYPE##_NONCONST                                    \
  ( boost::detail::multi_array::sub_array             ,  , INAME, ENAME,  \
    OBJECT )                                                              \
  MM_ARRAY_SCALAR_OP_##TYPE##_NONCONST                                    \
  ( boost::detail::multi_array::multi_array_view      , &, INAME, ENAME,  \
    OBJECT )                                                              \
  MM_ARRAY_SCALAR_OP_##TYPE##_CONST                                       \
  ( boost::detail::multi_array::const_sub_array       ,  , INAME, ENAME,  \
    OBJECT )                                                              \
  MM_ARRAY_SCALAR_OP_##TYPE##_CONST                                       \
  ( boost::detail::multi_array::const_multi_array_view, &, INAME, ENAME,  \
    OBJECT )

// Element-wise operations on arrays.
// f( array )
#define MM_ARRAY_OP_EXTERN( TYPE, NAME, OBJECT )                        \
  template< std::size_t N >                                             \
  boost::multi_array< typename OBJECT::result_type, N > NAME            \
  ( const TYPE< typename OBJECT::argument_type, N > & x )               \
  {                                                                     \
    mm_array_apply_extern< TYPE< typename OBJECT::argument_type, N >,   \
      OBJECT > f;                                                       \
    return f( x );                                                      \
  }

// array = f( array ); f( array )
#define MM_ARRAY_OP( NAME, OBJECT )                                     \
  MM_ARRAY_OP_EXTERN( boost::multi_array                                , \
                  NAME, OBJECT )                                        \
  MM_ARRAY_OP_EXTERN( boost::detail::multi_array::sub_array             , \
                  NAME, OBJECT )                                        \
  MM_ARRAY_OP_EXTERN( boost::detail::multi_array::multi_array_view      , \
                  NAME, OBJECT )                                        \
  MM_ARRAY_OP_EXTERN( boost::detail::multi_array::const_sub_array       , \
                  NAME, OBJECT )                                        \
  MM_ARRAY_OP_EXTERN( boost::detail::multi_array::const_multi_array_view, \
                  NAME, OBJECT )

// array1 = f( array1, array2 )
#define MM_ARRAY_ARRAY_OP_INTERN( LEFTTYPE, LEFTMODIFIER,               \
                                  RIGHTTYPE, NAME, OBJECT )             \
  template< typename T, std::size_t N >                                 \
  LEFTTYPE< T, N > LEFTMODIFIER NAME ( LEFTTYPE< T, N > LEFTMODIFIER x, \
                                       const RIGHTTYPE< T, N > &c )     \
  {                                                                     \
    mm_array_array_apply_intern< LEFTTYPE< T, N >, RIGHTTYPE< T, N >,   \
                                 OBJECT< T > > f;                       \
    return f( x, c );                                                   \
  }

// f( array1, array2 )
#define MM_ARRAY_ARRAY_OP_EXTERN( LEFTTYPE, RIGHTTYPE, NAME, OBJECT )   \
  template< typename T, std::size_t N >                                 \
  boost::multi_array< T, N > NAME ( const LEFTTYPE< T, N > &x,          \
                                    const RIGHTTYPE< T, N > &c )        \
  {                                                                     \
    mm_array_array_apply_extern< LEFTTYPE< T, N >, RIGHTTYPE< T, N >,   \
                                 OBJECT< T > > f;                       \
    return f( x, c );                                                   \
  }

// f( array1, array2 )
#define MM_ARRAY_ARRAY_OP_CONST_H2( LEFTTYPE, LEFTMODIFIER, RIGHTTYPE, \
                                    INAME, ENAME, OBJECT )             \
  MM_ARRAY_ARRAY_OP_EXTERN( LEFTTYPE, RIGHTTYPE, ENAME, OBJECT )

// array1 = f( array1, array2 ); f( array1, array2 )
#define MM_ARRAY_ARRAY_OP_NONCONST_H2( LEFTTYPE, LEFTMODIFIER, RIGHTTYPE,     \
                                       INAME, ENAME, OBJECT )                 \
  MM_ARRAY_ARRAY_OP_CONST_H2( LEFTTYPE, LEFTMODIFIER, RIGHTTYPE,              \
                              INAME, ENAME, OBJECT )                          \
  MM_ARRAY_ARRAY_OP_INTERN( LEFTTYPE, LEFTMODIFIER, RIGHTTYPE, INAME, OBJECT )

// array1 = f( array1, array2 ); f( array1, array2 )
#define MM_ARRAY_ARRAY_OP_H1( LEFTTYPE, PROTECT, LEFTMODIFIER, INAME, ENAME,  \
                              OBJECT )                                        \
  MM_ARRAY_ARRAY_OP_##PROTECT##_H2                                            \
  ( LEFTTYPE, LEFTMODIFIER,                                                   \
    boost::detail::multi_array::const_sub_array       , INAME, ENAME, OBJECT )\
  MM_ARRAY_ARRAY_OP_##PROTECT##_H2                                            \
  ( LEFTTYPE, LEFTMODIFIER,                                                   \
    boost::detail::multi_array::sub_array             , INAME, ENAME, OBJECT )\
  MM_ARRAY_ARRAY_OP_##PROTECT##_H2                                            \
  ( LEFTTYPE, LEFTMODIFIER,                                                   \
    boost::multi_array                                , INAME, ENAME, OBJECT )\
  MM_ARRAY_ARRAY_OP_##PROTECT##_H2                                            \
  ( LEFTTYPE, LEFTMODIFIER,                                                   \
    boost::detail::multi_array::const_multi_array_view, INAME, ENAME, OBJECT )\
  MM_ARRAY_ARRAY_OP_##PROTECT##_H2                                            \
  ( LEFTTYPE, LEFTMODIFIER,                                                   \
    boost::detail::multi_array::multi_array_view      , INAME, ENAME, OBJECT )

// array1 = f( array1, array2 ); f( array1, array2 )
#define MM_ARRAY_ARRAY_OP( INAME, ENAME, OBJECT )                           \
  MM_ARRAY_ARRAY_OP_H1( boost::detail::multi_array::const_sub_array       , \
                        CONST   ,  , INAME, ENAME, OBJECT )                 \
  MM_ARRAY_ARRAY_OP_H1( boost::detail::multi_array::sub_array             , \
                        NONCONST,  , INAME, ENAME, OBJECT )                 \
  MM_ARRAY_ARRAY_OP_H1( boost::multi_array                                , \
                        NONCONST, &, INAME, ENAME, OBJECT )                 \
  MM_ARRAY_ARRAY_OP_H1( boost::detail::multi_array::const_multi_array_view, \
                        CONST   , &, INAME, ENAME, OBJECT )                 \
  MM_ARRAY_ARRAY_OP_H1( boost::detail::multi_array::multi_array_view      , \
                        NONCONST, &, INAME, ENAME, OBJECT )


// Scalar multiplication
MM_ARRAY_SCALAR_OP( operator*=, operator*, std::multiplies, ASSOC    )
// Scalar division
MM_ARRAY_SCALAR_OP( operator/=, operator/, std::divides   , NONASSOC )
// Scalar addition
MM_ARRAY_SCALAR_OP( operator+=, operator+, std::plus      , ASSOC    )
// Scalar subtraction
MM_ARRAY_SCALAR_OP( operator-=, operator-, std::minus     , ASSOC    )

// Element-wise multiplication
MM_ARRAY_ARRAY_OP( operator*=, operator*, std::multiplies )
// Element-wise division
MM_ARRAY_ARRAY_OP( operator/=, operator/, std::divides    )
// Element-wise multiplication
MM_ARRAY_ARRAY_OP( operator+=, operator+, std::plus       )
// Element-wise division
MM_ARRAY_ARRAY_OP( operator-=, operator-, std::minus      )

#endif
