#ifndef MMVL_MM_PIXEL_CONVERT_HH
#define MMVL_MM_PIXEL_CONVERT_HH

#include "mm_rgba.h"
#include "mm_image.h" 
#ifndef NDEBUG
#include <limits>
#include <iostream>
#endif


namespace mimas{
  template<typename U, typename V>
  struct convert_trait{

    //find a way to round to the nearest value
    U convert(const V & src){
#ifndef NDEBUG
      std::numeric_limits<U> nl;
      assert(src <= nl.max());
#endif
      return (U)(src);
    }
    
#ifndef NDEBUG
    void debug(void){
      std::cerr<<"general trait"<<std::endl;
    }
#endif  
  };
  
  //template specialisation //
  
  template<typename U, typename V>
  struct convert_trait<mm_rgba<U>, mm_rgba<V> >{
    mm_rgba<U> convert(const mm_rgba<V> &src)
    {
      return mm_rgba<U>((U)(src.getRed()),(U)(src.getGreen()),(U)(src.getBlue()),(U)(src.getAlpha()));
    }

#ifndef NDEBUG
    void debug(void){
      std::cerr<<"color->color"<<std::endl;
    }
#endif  
  };

  template<typename U, typename V>
  struct convert_trait<U, mm_rgba<V> >{
    U convert(const mm_rgba<V> &src){
      double rwgt=0.3086;
      double gwgt=0.6094;
      double bwgt=0.0820;

      //!!!How to prevent overflow (i.e if U is unsigned char and value bigger than 255)
      //Again using the trait technique (see std::numeric_limits) but is it really neccesary? i.e. will we
      //have images with higher values than 255? ->think about filtered images.

      return (U)(rwgt*(double)src.getRed()+gwgt*(double)src.getGreen()+bwgt*(double)src.getBlue());
    }
#ifndef NDEBUG
    void debug(void){
      std::cerr<<"color->gray"<<std::endl;
    }
#endif  
  };

  //gray image to a color image
  template<typename U, typename V>
  struct convert_trait<mm_rgba<U>, V >{
    mm_rgba<U> convert(const V &src){
      return mm_rgba<U>((U)(src),(U)(src),(U)(src));
    }
#ifndef NDEBUG
    void debug(void){
      std::cerr<<"gray->color"<<std::endl;
    }
#endif
  };

}
#endif
