#include <cmath>

#include <qapplication.h>
#include <qlabel.h>
#include <qimage.h>
#include <qpixmap.h>

#ifndef MM_RGBA_H
#define MM_RGBA_H

namespace mimas {

  template <typename T>
    class mm_rgba
    {
    protected:
      T r,g,b,a;
		
    public:
		
      mm_rgba(void)
	{
	  r=g=b=a=T();
	}
	       
      template< class U >
		mm_rgba(const mm_rgba< U > &rhs)
		{
        r = (T)rhs.getRed();
        g = (T)rhs.getGreen();
        b = (T)rhs.getBlue();
        a = (T)rhs.getAlpha();
		}

      mm_rgba( T _r, T _g, T _b, T _a = T() ):
	r(_r), g(_g), b(_b), a(_a)
	{
	}
      
   explicit mm_rgba(const T &val)
	{
	  r=val;
	  g=val;
	  b=val;
	  a=T();
	}
      
      template< class U >
      mm_rgba& operator=(const mm_rgba< U > &rhs)
		{
        r = (T)rhs.getRed();
        g = (T)rhs.getGreen();
        b = (T)rhs.getBlue();
        a = (T)rhs.getAlpha();
        return *this;
		}

		mm_rgba& operator=(const T &rhs)
		{
			this->r=rhs;
			this->g=rhs;
			this->b=rhs;
			this->a=rhs;
			return *this;
		}


		//inline functions to avoid overhead of the call,
		//especially used for convert_trait
      inline T getRed(void) const
	{
	  return r;
	}

      inline T getGreen(void) const
	{
	  return g;
	}

      inline T getBlue(void) const
	{
	  return b;
	}
      
      inline T getAlpha(void) const
	{
	  return a;
	}
      

      void setRed(T val)
	{
	  r=val;
	}
      
      void setGreen(T val)
	{
	  g=val;
	}
      
      void setBlue(T val)
	{
	  b=val;
	}
      
      void setAlpha(T val)
	{
	  a=val;
	}
      
      void set(T val)
	{
	  r=g=b=a=val;
	}


      //mixes the 2 colors according to the ammount of transparency
      void overlay(T r_,T g_,T b_,T a_)
	{
	  //assumption that alpha take a value between 0 and 255
	  //should be corrected according to the type T
	  r = ((255 - a_)* r + a_ * r_)/255;
	  g = ((255 - a_)* g + a_ * g_)/255;
	  b = ((255 - a_)* b + a_ * b_)/255;
	}
      
		//overload casting to double
		operator double() const
      { return sqrt( ( (double)r * r +
                       (double)g * g +
                       (double)b * b ) / 3.0 ); }


      double getHue(void) const;///< returns number in the range 0-360
      double getSaturation(void) const; ///< returns saturation between 0 and 1.0
      
      T getGray(void) const;
    };
 
}

#include "mm_rgba.tcc"

#endif
 
