#ifndef MM_STEREO_H
#define MM_STEREO_H

#include <iostream> 
#include <cmath>
#include "mm_image.h"

namespace mimas {
template<typename T>
class mm_stereo : public mm_image <T>
{
//	protected:
//		mm_image<T> leftImage, rightImage;
		
    protected:
	mm_image<double> depthMap;
	
    public:
	 	bool fillgaps;
    	int xsearch, ysearch;
	int xwinsize, ywinsize;
   void stereoMap(mm_image<T> &leftImage, mm_image<T> &rightImage);
	void writeNormalisedDepthMapToFilePGM(char *fn);
	void writeNormalisedDepthMapMatrixToFile(char *fn);
        mm_stereo()
        {
	    xsearch=21;
	    ysearch=5;
	    xwinsize=5;
	    ywinsize=2;
		 fillgaps=false;
	}
	~mm_stereo()
	{
	}
};

template<typename T>
void mm_stereo<T>::stereoMap(mm_image<T> &leftImage, mm_image<T> &rightImage)
{
	// put in some error checking here like 
	// 1. whether image pairs have been loaded
	// 2. size of pairs etc
	
	int x,y,i,j;
	int offx, offy;


	if (!fillgaps)
	{
		depthMap.init(leftImage.getWidth(),leftImage.getHeight());
	}
	
	for (y=ysearch+ywinsize/2; y<leftImage.getHeight()-ysearch-ywinsize/2; y++)
	{
		for (x=xsearch+xwinsize/2; x<leftImage.getWidth()-xsearch-xwinsize/2; x++)
		{
			int xpoint, ypoint;
			double minEnergy, energy;
			int energy1, energy2, energy3;
			double depth;
			
			minEnergy=99999.0;
			
			for (i=x-xsearch/2; i<=x+xsearch/2; i++)
				for (j=y-ysearch/2; j<=y+ysearch/2; j++)
				{
					//energy=0;
					
					energy1=0;
					energy2=0;
					energy3=0;
					for (offx=-xwinsize/2; offx<=xwinsize/2; offx++)
						for (offy=-ywinsize/2; offy<=ywinsize/2; offy++)
						{
							//energy+=(int)sqrt( (leftImage.getPixel(x+offx,y+offy)-rightImage.getPixel(i+offx,j+offy)) *
							//			(leftImage.getPixel(x+offx,y+offy)-rightImage.getPixel(i+offx,j+offy)) );
										
							energy1+=( (leftImage.pixel(x+offx,y+offy)-rightImage.pixel(i+offx,j+offy)) *
										(leftImage.pixel(x+offx,y+offy)-rightImage.pixel(i+offx,j+offy)) );
							energy2+=leftImage.pixel(x+offx,y+offy)*leftImage.pixel(x+offx,y+offy);
							energy3+=rightImage.pixel(i+offx,j+offy)*rightImage.pixel(i+offx,j+offy);
							

						}

					energy=(double)energy1/sqrt((double)energy2*(double)energy3);
					
					if (energy<minEnergy)
					{
						minEnergy=energy;
						xpoint=i;
						ypoint=j;
					}

				}
				
			depth=(double)( ((x-xpoint)*(x-xpoint)) + ( (y-ypoint)*(y-ypoint) ));
			depth=sqrt(depth);
			//if ((int)minEnergy==0) depth=0.0;
			if (minEnergy<1e-16) depth=0.0;
					
//			int depthint=(int)depth;

			if (fillgaps)
			{
				if (depthMap.pixel(x,y)<1e-1)
              depthMap.pixel(x,y) = (double)depth;
			}
			else
           depthMap.pixel(x,y) = (double)depth;

//			depthMap.setPixel(x,y,(double)depthint);
//			depthMap.setPixel(x,y,(double)depth);
//			cout << (int)minEnergy << " ";
		}
	}
}


template<typename T>
void mm_stereo<T>::writeNormalisedDepthMapMatrixToFile(char *fn)
{
	mm_image<T> tempImage;
	int i,j, maxIntensity, minIntensity;


	tempImage.init(depthMap.getWidth(),depthMap.getHeight());
	for (i=0; i<depthMap.getWidth(); i++)
		for (j=0; j<depthMap.getHeight(); j++)
		{
			if (depthMap.pixel(i,j))
           tempImage.pixel(i,j) = (int)depthMap.pixel(i,j);
			else
           tempImage.pixel(i,j) = 0;
		}

	tempImage.writeToFilePGM(fn);
}


template<typename T>
void mm_stereo<T>::writeNormalisedDepthMapToFilePGM(char *fn)
{
	mm_image<T> tempImage;
	int i,j;
	double maxIntensity;

	tempImage.init(depthMap.getWidth(),depthMap.getHeight());

	maxIntensity=-999999.0;
	for (i=0; i<depthMap.getWidth(); i++)
		for (j=0; j<depthMap.getHeight(); j++)
		    if (maxIntensity<depthMap.pixel(i,j)) maxIntensity=depthMap.getpixel(i,j);

   tempImage = mm_convert_image< double, T >( depthMap );

	tempImage.maxgrayval=(T)(maxIntensity);
	tempImage.writeToFilePGM(fn);
}
}

#endif
