#ifndef MM_SVD_CORRESPONDENCE_TCC
#define MM_SVD_CORRESPONDENCE_TCC

#include <boost/numeric/ublas/matrix.hpp>
#include "mm_linalg.h"

namespace mimas {
template< class InputIter1, class SimilarityMeasure >
template< class InputIter2, class OutputIter1, class OutputIter2 >
void mm_svd_correspondence< InputIter1, SimilarityMeasure >::match
     ( InputIter2 begin2, int size2, OutputIter1 outputIndexed,
       OutputIter2 outputLost ) throw (mm_exception)
{
  if ( size1 > 0 && size2 > 0 ) {

    Matrix m( size1, size2 );
   
    InputIter1 trackingI = begin1;
    for ( int row = 0; row<(signed)m.size1(); row++ ) {
      InputIter2 trackingJ = begin2;
      for ( int col=0; col<(signed)m.size2(); col++ ) {
        m( row, col ) = similarityMeasure( trackingI->second, *trackingJ );
        trackingJ++;
      };
      trackingI++;
    };
    
    Matrix u, vt;
    mm_gesdd< double > gesdd;

    // The singular values are ignored. Only the left- and right-hand vectors
    // are of interest.
    gesdd( m, &u, &vt );
    
    // Replace singular values with '1.0'.
    // IdentityMatrix identity( size1, size2 );
    boost::numeric::ublas::identity_matrix< double > identity( size1, size2 );
    Matrix a( identity );
    // Nested function calls are not working in release mode for some reason.
    Matrix
	tmp( prod( u, identity ) ),
	correspondence( prod( tmp, vt ) );
    
    std::vector< std::pair< int, double > > maxCol, maxRow;
    maxRow.resize( size2, std::pair< int, double >( -1, 0.0 ) );
    maxCol.resize( size1, std::pair< int, double >( -1, 0.0 ) );
    
    for ( int i=0; i<size1; i++ )
      for ( int j=0; j<size2;j++ ) {
        double value = fabs( correspondence( i, j ) );
        if ( value > maxRow.at( j ).second )
          maxRow.at( j ) = std::pair< int, double >( i, value );
        if ( value > maxCol.at( i ).second )
          maxCol.at( i ) = std::pair< int, double >( j, value );
      };
    
    {
      std::vector< bool > found( size2 );

      InputIter1 trackingI = begin1;
      
      for ( int i=0; i<size1; i++, trackingI++ ) {
        
        int j = maxCol.at( i ).first;
        if ( j != -1 )
          if ( maxRow.at( j ).first == i ) {
            *outputIndexed++ = typename InputIter1::value_type
              ( trackingI->first, *( begin2 + j ) );
            assert( found[ j ] == false );
            found[ j ] = true;
          };
      };

      InputIter2 lostJ = begin2;
      for ( int j=0; j<size2; j++ ) {
        if ( !found[ j ] )
          *outputLost++ = (*lostJ);
        lostJ++;
      };

    };
  };
 }
}
#endif
