//
// Video4Linux interface for grabbers and webcams
//


#ifndef MM_VIDEO4LINUX
#define MM_VIDEO4LINUX

#include <iostream>
#include <cmath>
#include <cstdio>
#include "mm_exception.h"
#include "mm_image.h"
#include "mm_rgba.h"

/* for the v4l stuff: */
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <linux/videodev.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

namespace mimas {


/** Video4Linux interface
    IMPORTANT COPYRIGHT NOTE: The three functions for YUV420p -> RGB conversion,
    namely limit(), copy420Block() and yuv420p_to_rgb() is taken from
    linux kernel linux-2.4.22/drivers/usb/ov511.c. Note that these functions
    are GPL-ed not LGPL-ed. Can anyone advice if this needs rewriting?
    @author Bala Amavasai (bala@amavasai.org)
    @date Thu Oct 16 12:04:45 BST 2003
    $Header: /cvs/mimas/include/mm_video4linux.h,v 1.9 2004/12/09 23:27:19 engjw Exp $ */
class mm_video4linux //: public  mm_object
{
	private:
		char VIDEO_DEVICE[255]; ///< the video device
		bool initialised;
		int fd; ///< file descriptor for video device
		struct video_mmap v_map; ///< the video map
		struct video_picture pic; ///< the picture property

		int chan_number, chan_type, chan_norm; ///< channel information
		int width, height; ///< width and height of image frame
		int palette; ///< v4l palette
		int depth; ///< image depth
		int brightness, hue, colour, contrast, whiteness; ///< picture properties

	protected:
		void *map;
		int limit(int x); ///< helper for yuv420p_to_rgb()
		void copy420Block (int yTL, int yTR, int yBL, int yBR, int u, int v, int rowPixels, unsigned char *rgb, int bits); ///< conversion routine, see yuv420p_to_rgb()
		int yuv420p_to_rgb (unsigned char *out, unsigned char *in, int width, int height, int bits); ///< convert YUV420P data to RGB data
      void capture(void) throw(mm_exception); ///< the low level function that performs the capture. Should be called by grab

	public:
		mm_video4linux(void); ///< constructor
		int init(void); ///< initialise the grabber (optional)
      virtual void grab(mm_image< unsigned char > &imagein) throw (mm_exception); ///< grab an image and store it as grayscale
      virtual void grab(mm_image< mm_rgba< unsigned char > > &imagein) throw (mm_exception); ///< grab an image and store it in colour
      void setDevice(char *devname) throw (mm_exception); ///< set the device. Default is /dev/video0
		void setChannelNumber(int num) throw (mm_exception); ///< set the device channel number
      void setChannelInputType(int num) throw (mm_exception); ///< set the device input type
      void setChannelVideoMode(int num) throw (mm_exception); ///< set the device channel mode
		void setPalette(int p) throw (mm_exception); ///< set palette
      void setDepth(int d) throw (mm_exception); ///< set depth
		void setBrightnessLevel(int b); ///< set brightness
		void setHueLevel(int h); ///< set hue
		void setColourLevel(int c); ///< set colour saturation
		void setContrastLevel(int c); ///< set contrast
		void setWhitenessLevel(int w); ///< set whiteness level
		void setWidth(int w) throw (mm_exception); ///< set frame width
		void setHeight(int h) throw (mm_exception); ///< set frame height
		int getHeight(void); ///< get height of grabbed frame
		int getWidth(void); ///< get width of grabbed frame

};

}
#endif
