#include <boost/multi_array.hpp>
#include "mm_image_fileoutput.h"
#include "mm_rgba.h"

using namespace boost;

namespace mimas {

class mm_ostream_wrapper: public QIODevice {
 public:
  mm_ostream_wrapper( std::basic_ostream<char> &_stream ):
    stream( &_stream ) {}
  virtual ~mm_ostream_wrapper(void) {}
  virtual bool open(int) { return false; }
  virtual void close(void) {}
  virtual void flush(void) {}
  virtual Offset size(void) const { return 0; }
  virtual Q_LONG readBlock( char*, Q_ULONG ) { return -1; }
  virtual Q_LONG writeBlock( const char* data, Q_ULONG len );
  virtual int putch(int);
  virtual int ungetch(int) { return -1; }
  virtual int getch(void) { return -1; }
  virtual bool at( Offset pos )
  { stream->seekp( pos, std::ios::beg ); return *stream; }
 protected:
  std::basic_ostream<char> *stream;
};

Q_LONG mm_ostream_wrapper::writeBlock( const char* data, Q_ULONG len )
{
  stream->write( data, len );
  Q_LONG retVal;
  if ( *stream )
    retVal = len;
  else
    retVal = -1;
  return retVal;
}

int mm_ostream_wrapper::putch( int ch )
{
  stream->put( ch );
  return *stream ? ch : -1;
}

static QImage mm_image_to_qimage( const mm_image< unsigned char > &image )
{
  QImage retVal( image.getWidth(), image.getHeight(), 8, 256 );
  for ( int i=0; i<256; i++ )
    retVal.setColor( i, qRgb( i, i, i ) );
    // Line-wise copying required, as QImage does 32 bit memory alignment!
  for ( int y=0; y<image.getHeight(); y++ )
    copy_n( image.rawData()[y].begin(), image.getWidth(),
            retVal.scanLine( y ) );
  return retVal;
}

static QImage mm_rgba_image_to_qimage( const mm_image< mm_rgba< unsigned char > > &image )
{
  QImage img( image.getWidth(), image.getHeight(), 32 );
  assert( sizeof( mm_rgba< unsigned char > ) == 4 );
  // Line-wise copying required, as QImage does 32 bit memory alignment!
  for ( int y=0; y<image.getHeight(); y++ )
    copy_n( image.rawData()[y].begin(), image.getWidth(),
            (mm_rgba< unsigned char > *)img.scanLine( y ) );
  return img.swapRGB();
}

void mm_image_fileoutput< unsigned char >::write
   ( const mm_image< unsigned char > &image )
{
  mm_ostream_wrapper ostreamWrapper( *stream );
  // Default file type is pgm.
  if ( mm_image_fileoutput_fileType == NULL )
    mm_image_fileoutput_fileType = "PGM";
  QImageIO io( &ostreamWrapper, mm_image_fileoutput_fileType );
  mm_image_fileoutput_fileType = NULL;
  io.setImage( mm_image_to_qimage( image ) );
  if ( !io.write() )
    stream->setstate( std::ios_base::failbit );
}

void mm_image_fileoutput< mm_rgba< unsigned char > >::write
   ( const mm_image< mm_rgba< unsigned char > > &image )
{
  mm_ostream_wrapper ostreamWrapper( *stream );
  // Default file type is ppm.
  if ( mm_image_fileoutput_fileType == NULL )
    mm_image_fileoutput_fileType = "PPM";
  QImageIO io( &ostreamWrapper, mm_image_fileoutput_fileType );
  mm_image_fileoutput_fileType = NULL;
  io.setImage( mm_rgba_image_to_qimage( image ) );
  if ( !io.write() )
    stream->setstate( std::ios_base::failbit );
}

/// File-format of next image to be written.
const char *mm_image_fileoutput_fileType = NULL;

};
