#include <boost/multi_array.hpp>
#include "mm_image_qtoutput.h"

using namespace boost;

namespace mimas {

static QPixmap mm_image_to_qpixmap
 ( const mm_image< unsigned char > &image )
{
  QImage img( image.getWidth(), image.getHeight(), 8, 256 );
  for ( int i=0; i<256; i++ )
    img.setColor( i, qRgb( i, i, i ) );

  // Line-wise copying required, as QImage does 32 bit memory alignment!
  for ( int y=0; y<image.getHeight(); y++ )
    copy_n( image.rawData()[y].begin(), image.getWidth(), img.scanLine( y ) );
  QPixmap retVal;
  retVal.convertFromImage(img);
  return retVal;
}

static QPixmap mm_image_to_qpixmap
 ( const mm_image< mm_rgba< unsigned char > > &image )
{
  QImage img( image.getWidth(), image.getHeight(), 32 );
  assert( sizeof( mm_rgba< unsigned char > ) == 4 );
  // Line-wise copying required, as QImage does 32 bit memory alignment!
  for ( int y=0; y<image.getHeight(); y++ )
    copy_n( image.rawData()[y].begin(), image.getWidth(),
            (mm_rgba< unsigned char > *)img.scanLine( y ) );
  
  QPixmap retVal;
  retVal.convertFromImage( img.swapRGB() );
  return retVal;
}

void mm_image_qtoutput< unsigned char >::write
   ( const mm_image< unsigned char > &image )
{
  writePixmap( mm_image_to_qpixmap( image ) );
}

void mm_image_qtoutput< mm_rgba< unsigned char > >::write
   ( const mm_image< mm_rgba< unsigned char > > &image )
{
  writePixmap( QPixmap( mm_image_to_qpixmap( image ) ) );
}

};
