#include <gst/gst.h>
#include <gst/video/video.h>
#include <gst/video/videosink.h>
#include <string.h>
#include <sys/sem.h>
#include <sys/shm.h>

typedef struct _GstMimasImageSink GstMimasImageSink;

struct _GstMimasImageSink {
  GstVideoSink videosink;
  GstPad *sinkpad,*srcpad;
  GstClockTime time;
  gint bpp, depth, endianness, red_mask, green_mask, blue_mask;
  gdouble framerate;
  int shmid;
  int semid;
  int shmkey;
  int semkey;
  int shmsize;
  void *shm;
};

typedef struct _GstMimasImageSinkClass GstMimasImageSinkClass;

struct _GstMimasImageSinkClass {
  GstVideoSinkClass parent_class;
};

#define GST_TYPE_MIMASIMAGESINK \
  (gst_mimasimagesink_get_type())
#define GST_MIMASIMAGESINK(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_MIMASIMAGESINK,GstMimasImageSink))
#define GST_MIMASIMAGESINK_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_MIMASIMAGESINK,GstMimasImageSinkClass))
#define GST_IS_MIMASIMAGESINK(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_MIMASIMAGESINK))
#define GST_IS_MIMASIMAGESINK_CLASS(obj) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_MIMASIMAGESINK))

static GstElementDetails mimasimagesink_details =
GST_ELEMENT_DETAILS( "Mimas image sink",
                     "Sink/Video",
                     "Feed images to mimas application",
                     "Jan Wedekind <jan@wedesoft.de>" );

enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
  ARG_SHMKEY,
  ARG_SEMKEY
  /* FILL ME */
};

#define SHMKEYDEFAULT 333
#define SEMKEYDEFAULT 333

GstStaticPadTemplate sink_template =
  GST_STATIC_PAD_TEMPLATE ( "sink",
                            GST_PAD_SINK,
                            GST_PAD_ALWAYS,
                            GST_STATIC_CAPS
                            ( GST_VIDEO_CAPS_RGB ) );

// Videotestsrc: RGB, RGBx
// QImage: BGR
// dvdec: RGB, ARGB

static void gst_mimasimagesink_class_init (GstMimasImageSinkClass * klass);
static void gst_mimasimagesink_init (GstMimasImageSink * mimasimagesink);
static GstPadLinkReturn
gst_mimasimagesink_sink_link (GstPad * pad, const GstCaps * caps);
static GstCaps *gst_mimasimagesink_fixate (GstPad * pad, const GstCaps * caps);
static GstBuffer *
gst_mimasimagesink_buffer_alloc (GstPad * pad, guint64 offset, guint size);
static void gst_mimasimagesink_buffer_free (GstBuffer * buffer);
static void gst_mimasimagesink_chain (GstPad * pad, GstData * _data);
static void gst_mimasimagesink_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_mimasimagesink_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstElementStateReturn gst_mimasimagesink_change_state
   (GstElement * element);

static GstElementClass *parent_class = NULL;

GType gst_mimasimagesink_get_type(void)
{
  static GType mimasimagesink_type = 0;

  if (!mimasimagesink_type) {
    static const GTypeInfo mimasimagesink_info = {
      sizeof (GstMimasImageSinkClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_mimasimagesink_class_init,
      NULL,
      NULL,
      sizeof (GstMimasImageSink),
      0,
      (GInstanceInitFunc) gst_mimasimagesink_init,
    };
    mimasimagesink_type =
      g_type_register_static (GST_TYPE_VIDEOSINK, "GstMimasImageSink",
                              &mimasimagesink_info, 0 );
  }
  return mimasimagesink_type;
}

static void
gst_mimasimagesink_class_init (GstMimasImageSinkClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  parent_class = g_type_class_ref (GST_TYPE_VIDEOSINK);

  g_object_class_install_property
    ( G_OBJECT_CLASS( klass ), ARG_SHMKEY,
      g_param_spec_int( "shmkey", "shared memory key",
                        "key of shared memory block",
                        1, G_MAXINT, SHMKEYDEFAULT, G_PARAM_READWRITE ) );
  g_object_class_install_property
    ( G_OBJECT_CLASS( klass ), ARG_SEMKEY,
      g_param_spec_int( "semkey", "semaphore key",
                        "key of synchronisation semaphore",
                        1, G_MAXINT, SEMKEYDEFAULT, G_PARAM_READWRITE ) );

  gobject_class->set_property = gst_mimasimagesink_set_property;
  gobject_class->get_property = gst_mimasimagesink_get_property;

  gstelement_class->change_state = gst_mimasimagesink_change_state;
  gst_element_class_set_details (gstelement_class, &mimasimagesink_details);
  gst_element_class_add_pad_template( gstelement_class,
                                      gst_static_pad_template_get
                                      ( &sink_template ) );
}

static void
gst_mimasimagesink_init (GstMimasImageSink * mimasimagesink)
{
  GST_VIDEOSINK_PAD( mimasimagesink ) =
      gst_pad_new_from_template( gst_static_pad_template_get (&sink_template),
                                 "sink");
  gst_pad_set_chain_function( GST_VIDEOSINK_PAD(mimasimagesink),
                              gst_mimasimagesink_chain );
  gst_pad_set_link_function( GST_VIDEOSINK_PAD(mimasimagesink),
                             gst_mimasimagesink_sink_link );
  gst_pad_set_fixate_function( GST_VIDEOSINK_PAD(mimasimagesink),
                               gst_mimasimagesink_fixate);
  gst_pad_set_bufferalloc_function( GST_VIDEOSINK_PAD(mimasimagesink),
                                    gst_mimasimagesink_buffer_alloc );
  gst_element_add_pad( GST_ELEMENT(mimasimagesink),
                       GST_VIDEOSINK_PAD(mimasimagesink) );

  mimasimagesink->shmkey = SHMKEYDEFAULT;
  mimasimagesink->semkey = SEMKEYDEFAULT;
  mimasimagesink->shmid = -1;
  mimasimagesink->semid = -1;
  mimasimagesink->shm = NULL;
  mimasimagesink->shmsize = 0;
  GST_FLAG_SET(mimasimagesink, GST_ELEMENT_THREAD_SUGGESTED);
}

static void
gst_mimasimagesink_chain (GstPad * pad, GstData * _data)
{
  GstBuffer *buf = GST_BUFFER (_data);
  GstMimasImageSink *mimasimagesink =
    GST_MIMASIMAGESINK (gst_pad_get_parent (pad));

  if (GST_IS_EVENT(_data) ) {
    gst_pad_event_default(pad, GST_EVENT(_data));
    return;
  }

  g_return_if_fail (pad != NULL);
  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);

  g_return_if_fail (mimasimagesink != NULL);
  g_return_if_fail (GST_IS_MIMASIMAGESINK (mimasimagesink));

  if ( GST_BUFFER_TIMESTAMP_IS_VALID( buf ) ) {
    mimasimagesink->time = GST_BUFFER_TIMESTAMP( buf );
  }
  if ( GST_VIDEOSINK_CLOCK( mimasimagesink ) ) {
    gst_element_wait( GST_ELEMENT( mimasimagesink ), mimasimagesink->time );
  }

  /* g_print("%dx%d=%d buf=%d fps=%f\n",GST_VIDEOSINK_WIDTH(mimasimagesink),
          GST_VIDEOSINK_HEIGHT(mimasimagesink),
          GST_VIDEOSINK_WIDTH(mimasimagesink)*
          GST_VIDEOSINK_HEIGHT(mimasimagesink),
          GST_BUFFER_SIZE( buf ),
          mimasimagesink->framerate); */

  if ( mimasimagesink->semid >= 0 ) {
    if ( mimasimagesink->shm == NULL ) {
      /* Allocate space for RGBA buffer with size-information. */
      mimasimagesink->shmsize =
        3 * GST_VIDEOSINK_WIDTH(mimasimagesink) *
        GST_VIDEOSINK_HEIGHT(mimasimagesink) +
        2 * sizeof(int);
      mimasimagesink->shmid = shmget( mimasimagesink->shmkey,
                                      mimasimagesink->shmsize,
                                      IPC_CREAT | 0666 );
      if ( mimasimagesink->shmid >= 0 ) {
        mimasimagesink->shm = shmat( mimasimagesink->shmid, NULL, 0 );
        if ( mimasimagesink->shm == NULL )
          shmctl( mimasimagesink->shmid, IPC_RMID, NULL );
        else
          g_print("mimasimagesink: Allocated shared memory.\n");
      };
    };
    
    if ( mimasimagesink->shm != NULL ) {
      if ( GST_BUFFER_SIZE( buf ) > ( mimasimagesink->shmsize - 8 ) )
        GST_WARNING("mimasimagesink: Size of incoming image to big!\n");
      else {
        int *dim = (int *)mimasimagesink->shm;
        dim[ 0 ] = GST_VIDEOSINK_WIDTH(mimasimagesink);
        dim[ 1 ] = GST_VIDEOSINK_HEIGHT(mimasimagesink);
        if ( GST_BUFFER_FREE_DATA_FUNC(buf) !=
             gst_mimasimagesink_buffer_free ) {
          memcpy( mimasimagesink->shm + 8, GST_BUFFER_DATA( buf ),
                  GST_BUFFER_SIZE( buf ) );
        };
        /* Increase read semaphore. */
        struct sembuf sbuf = { 0, +1, 0 };
        semop( mimasimagesink->semid, &sbuf, 1 );
      };
    };
  };

  if ( !GST_BUFFER_TIMESTAMP_IS_VALID(buf) && mimasimagesink->framerate>0) {
    mimasimagesink->time += GST_SECOND / mimasimagesink->framerate;
  };
  gst_buffer_unref( buf );
}

static GstPadLinkReturn
gst_mimasimagesink_sink_link (GstPad * pad, const GstCaps * caps)
{
  GstMimasImageSink *mimasimagesink =
    GST_MIMASIMAGESINK( gst_pad_get_parent( pad ) );
  GstStructure *structure = gst_caps_get_structure( caps, 0 );
  GstPadLinkReturn retVal;
  gboolean ret = TRUE;
  ret = ret && gst_structure_get_int( structure, "bpp" ,
                                      &mimasimagesink->bpp );
  ret = ret && gst_structure_get_int( structure, "depth" ,
                                      &mimasimagesink->depth );
  ret = ret && gst_structure_get_int( structure, "endianness" ,
                                      &mimasimagesink->endianness );
  ret = ret && gst_structure_get_int( structure, "red_mask" ,
                                      &mimasimagesink->red_mask );
  ret = ret && gst_structure_get_int( structure, "green_mask" ,
                                      &mimasimagesink->green_mask );
  ret = ret && gst_structure_get_int( structure, "blue_mask" ,
                                      &mimasimagesink->blue_mask );
  ret = ret && gst_structure_get_int( structure, "width" ,
                                      &GST_VIDEOSINK_WIDTH(mimasimagesink) );
  ret = ret && gst_structure_get_int( structure, "height",
                                      &GST_VIDEOSINK_HEIGHT(mimasimagesink) );
  ret = ret && gst_structure_get_double( structure, "framerate",
                                         &mimasimagesink->framerate );
  /* g_print( "%dx%d fps=%f endian=%d red=%x green=%x blue=%x: ",
           (int)GST_VIDEOSINK_WIDTH(mimasimagesink),
           (int)GST_VIDEOSINK_HEIGHT(mimasimagesink),
           mimasimagesink->framerate,
           mimasimagesink->endianness,
           mimasimagesink->red_mask, 
           mimasimagesink->green_mask,
           mimasimagesink->blue_mask ); */
  if ( ret ) {
    GST_WARNING("mimasimagesink: accepting caps\n");
    retVal = GST_PAD_LINK_OK;
  } else {
    GST_WARNING("mimasimagesink: error getting video information.zn");
    retVal = GST_PAD_LINK_REFUSED;
  };
  return retVal;
}

static GstCaps *
gst_mimasimagesink_fixate (GstPad * pad, const GstCaps * caps)
{
  GstStructure *structure;
  GstCaps *newcaps;

  if (gst_caps_get_size (caps) > 1)
    return NULL;

  newcaps = gst_caps_copy (caps);
  structure = gst_caps_get_structure (newcaps, 0);

  if (gst_caps_structure_fixate_field_nearest_int (structure, "width", 320)) {
    return newcaps;
  }
  if (gst_caps_structure_fixate_field_nearest_int (structure, "height", 240)) {
    return newcaps;
  }
  if (gst_caps_structure_fixate_field_nearest_double (structure, "framerate",
                                                      5.0)) {
    return newcaps;
  }

  gst_caps_free (newcaps);
  return NULL;
}

static void
gst_mimasimagesink_set_property( GObject * object, guint prop_id,
                                 const GValue * value, GParamSpec * pspec )
{
  GstMimasImageSink *mimasimagesink = GST_MIMASIMAGESINK (object);

  g_return_if_fail (GST_IS_MIMASIMAGESINK (object));

  switch (prop_id) {
  case ARG_SHMKEY:
    mimasimagesink->shmkey = g_value_get_int(value);
    g_print ("mimasimagesink: set shmkey to %d\n", mimasimagesink->shmkey );
    break;
  case ARG_SEMKEY:
    mimasimagesink->semkey = g_value_get_int(value);
    g_print ("mimasimagesink: set semkey to %d\n", mimasimagesink->semkey );
    break;
  default:
    break;
  }
}

static GstBuffer *
gst_mimasimagesink_buffer_alloc( GstPad * pad, guint64 offset, guint size )
{
  GstMimasImageSink *mimasimagesink =
    GST_MIMASIMAGESINK (gst_pad_get_parent( pad ) );
  GstBuffer *buffer;

  if ( mimasimagesink->semid >= 0 ) {
    struct sembuf sbuf = { 1, -1, 0 };
    /* Decrease write semaphore. */
    semop( mimasimagesink->semid, &sbuf, 1 );
    if ( mimasimagesink->shm != NULL ) {
      buffer = gst_buffer_new ();
      /* No private information needed. */
      GST_BUFFER_PRIVATE (buffer) = NULL;
      GST_BUFFER_DATA (buffer) = mimasimagesink->shm + 8;
      GST_BUFFER_FREE_DATA_FUNC (buffer) = gst_mimasimagesink_buffer_free;
      GST_BUFFER_SIZE (buffer) = mimasimagesink->shmsize - 8;
      return buffer;
    } else
      return NULL;
  } else
    return NULL;
}

static void
gst_mimasimagesink_buffer_free( GstBuffer * buffer )
{
}

static void
gst_mimasimagesink_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  GstMimasImageSink *mimasimagesink;

  /* It's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_MIMASIMAGESINK (object));
  mimasimagesink = GST_MIMASIMAGESINK (object);

  switch (prop_id) {
  case ARG_SHMKEY:
    g_value_set_int( value, mimasimagesink->shmkey );
    break;
  case ARG_SEMKEY:
    g_value_set_int( value, mimasimagesink->semkey );
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static GstElementStateReturn
gst_mimasimagesink_change_state (GstElement * element)
{
  GstMimasImageSink *mimasimagesink = GST_MIMASIMAGESINK (element);

  switch (GST_STATE_TRANSITION (element)) {
    case GST_STATE_NULL_TO_READY:
      mimasimagesink->semid =
        semget( mimasimagesink->semkey, 2, IPC_CREAT | 0666 );
      if ( mimasimagesink->semid < 0 )
        return GST_STATE_FAILURE;
      else {
        struct sembuf sbuf = { 1, +1, 0 };
        g_print("mimasimagesink: Allocated semaphore.\n");
        /* Increase write semaphore. */
        semop( mimasimagesink->semid, &sbuf, 1 );
      };
      break;
    case GST_STATE_READY_TO_PAUSED:
      mimasimagesink->time = 0;
      break;
    case GST_STATE_PAUSED_TO_PLAYING:
      break;
    case GST_STATE_PLAYING_TO_PAUSED:
      break;
    case GST_STATE_PAUSED_TO_READY:
      break;
    case GST_STATE_READY_TO_NULL:
      if ( mimasimagesink->shm != NULL ) {
        g_print("mimasimagesink: Removing shared memory.\n");
        shmdt( mimasimagesink->shm );
        mimasimagesink->shm = NULL;
        shmctl( mimasimagesink->shmid, IPC_RMID, NULL );
        mimasimagesink->shmid = -1;
      };
      if ( mimasimagesink->semid >= 0 ) {
        g_print("mimasimagesink: Removing semaphore.\n");
        semctl( mimasimagesink->semid, 0, IPC_RMID );
        mimasimagesink->semid = -1;
      };
      break;
  }

  if ( GST_ELEMENT_CLASS(parent_class)->change_state )
    return GST_ELEMENT_CLASS(parent_class)->change_state (element);
  return GST_STATE_SUCCESS;
}

static gboolean
plugin_init (GstPlugin * plugin)
{
  /* Loading the library containing GstVideoSink, our parent object */
  if (!gst_library_load ("gstvideo")) {
    GST_WARNING("mimasimagesink: Could not load gstvideo library.");
    return FALSE;
  };

  if (!gst_element_register (plugin, "mimasimagesink", GST_RANK_MARGINAL,
                             GST_TYPE_MIMASIMAGESINK)) {
    GST_WARNING("mimasimagesink: Could not register element.");
    return FALSE;
  };

  return TRUE;
}

GST_PLUGIN_DEFINE( GST_VERSION_MAJOR,
                   GST_VERSION_MINOR,
                   "mimasimagesink",
                   "Feed images to MIMAS vision library",
                   plugin_init, "0.1", "GPL", "mimas",
                   "http://www.shu.ac.uk/mmvl/mimas/" );
