#include "mm_linalg.h"

namespace mimas {

#ifdef HAVE_LIBLAPACK

int getrf_< double >::operator()( integer *m, integer *n, double *a,
                                  integer *lda, integer *ipiv, integer *info )
  const
{
  return dgetrf_( m, n, a, lda, ipiv, info );
}

int getrf_< float >::operator()( integer *m, integer *n, float *a,
                                 integer *lda, integer *ipiv, integer *info )
  const
{
  return sgetrf_( m, n, a, lda, ipiv, info );
}

int getri_< double >::operator()( integer *n, double *a, integer *lda,
                                  integer *ipiv, double *work,
                                  integer *lwork, integer *info ) const
{
  return dgetri_( n, a, lda, ipiv, work, lwork, info );
}

int getri_< float >::operator()( integer *n, float *a, integer *lda,
                                 integer *ipiv, float *work,
                                 integer *lwork, integer *info ) const
{
  return sgetri_( n, a, lda, ipiv, work, lwork, info );
}

int gglse_< double >::operator()( integer *m, integer *n, integer *p,
                                  double *a, integer *lda, double *b,
                                  integer *ldb, double *c__, double *d__,
                                  double *x, double *work, integer *lwork,
                                  integer *info )
{
  return dgglse_( m, n, p, a, lda, b, ldb, c__, d__, x, work, lwork, info );
}

int gglse_< float >::operator()( integer *m, integer *n, integer *p,
                                 float *a, integer *lda, float *b,
                                 integer *ldb, float *c__, float *d__,
                                 float *x, float *work, integer *lwork,
                                 integer *info )
{
  return sgglse_( m, n, p, a, lda, b, ldb, c__, d__, x, work, lwork, info );
}

int gels_< double >::operator()( char *trans, integer *m, integer *n,
                                 integer *nrhs, double *a, integer *lda,
                                 double *b, integer *ldb, double *work,
                                 integer *lwork, integer *info ) const
{
  return dgels_( trans, m, n, nrhs, a, lda, b, ldb, work, lwork, info );
}

int gels_< float >::operator()( char *trans, integer *m, integer *n,
                                integer *nrhs, float *a, integer *lda,
                                float *b, integer *ldb, float *work,
                                integer *lwork, integer *info ) const
{
  return sgels_( trans, m, n, nrhs, a, lda, b, ldb, work, lwork, info );
}

int gesvd_< double >::operator()( char *jobu, char *jobvt, integer *m,
                                  integer *n, double *a, integer *lda,
                                  double *s, double *u, integer *ldu,
                                  double *vt, integer *ldvt, double *work,
                                  integer *lwork, integer *info) const
{
  return dgesvd_( jobu, jobvt, m, n, a, lda, s, u, ldu, vt, ldvt, work, lwork,
                  info );
}

int gesvd_< float >::operator()( char *jobu, char *jobvt, integer *m,
                                 integer *n, float *a, integer *lda,
                                 float *s, float *u, integer *ldu,
                                 float *vt, integer *ldvt, float *work,
                                 integer *lwork, integer *info) const
{
  return sgesvd_( jobu, jobvt, m, n, a, lda, s, u, ldu, vt, ldvt, work, lwork,
                  info );
}

int gesdd_< double >::operator()( char *jobz, integer *m, integer *n,
                                  double *a, integer *lda, double *s,
                                  double *u, integer *ldu, double *vt,
                                  integer *ldvt, double *work, integer *lwork,
                                  integer *iwork, integer *info )
{
  return dgesdd_( jobz, m, n, a, lda, s, u, ldu, vt, ldvt, work, lwork,
                  iwork, info );
}

int gesdd_< float >::operator()( char *jobz, integer *m, integer *n,
                                 float *a, integer *lda, float *s,
                                 float *u, integer *ldu, float *vt,
                                 integer *ldvt, float *work, integer *lwork,
                                 integer *iwork, integer *info )
{
  return sgesdd_( jobz, m, n, a, lda, s, u, ldu, vt, ldvt, work, lwork,
                  iwork, info );
}

#endif

};
