#include "mm_linalg.h"
#include "mm_opticflow_diff_tool.h"

#ifdef HAVE_LIBLAPACK

namespace mimas {
using namespace std;
using namespace boost::numeric;

mm_opticflow_diff_tool::mm_opticflow_diff_tool(void)
{
	windowdist=13;
	//variance=2.25; // s.d. =1.5
	variance=1.0;
}


mm_opticflow_diff_tool::~mm_opticflow_diff_tool(void)
{
}


void mm_opticflow_diff_tool::smoothGaussian(mm_image<double>& imagein, double variance)
{
	mm_image<double> imageout;
	double divcoef;
	const int fsize=11;
	
	vector < vector<float> > f(fsize,fsize);

	divcoef=0.0;
	for (int j=-fsize/2; j<=fsize/2; j++)
	{
		for (int i=-fsize/2; i<=fsize/2; i++)
		{
			f[i+fsize/2][j+fsize/2]=exp( -(j*j+i*i)/(2*variance) );
			divcoef+=(f[i+fsize/2][j+fsize/2]);
		//	cout << f[i+fsize/2][j+fsize/2] << "\t" ;
		}
		//cout << endl;
	}


	convolve(imagein,imageout,f, (float)divcoef, (float)0.0);
	imagein=imageout;
}



void mm_opticflow_diff_tool::findFlow(mm_image<int> &image_t, mm_image<int> &image_dt)
{
	MMERROR( (image_t.getWidth()==image_dt.getWidth()) &&
            (image_t.getHeight()!=image_dt.getHeight()), mm_exception, ,
            "mm_opticflow_diff_tool::findFlow - image sizes are inconsistent!" );

	mm_image<double>
     it( mm_convert_image< double >( image_t ) ),
     idt( mm_convert_image< double >( image_dt ) );

	//findFlowLucasKanade(it,idt);

}


void mm_opticflow_diff_tool::findFlow(vector < mm_image<double> > &imagein)
{
   MMERROR( imagein.size()==5, mm_exception, ,
            "mm_opticflow_diff_tool::findFlowLucasKanade - need 5 images in "
            "vector list!" );

	//for (int d=0; d<imagein.size(); d++)
	//	imagein[d].display();

	mm_image<double> curframe;

	curframe=imagein[2];
	//smoothGaussian(curframe,variance);

	//curframe.display();

	mm_image<double> gradx, grady, gradt;
	double gx,gy, gt;
	int gsize=5;
	double D5[5]={-1.0/12.0, 8.0/12.0, 0.0/12.0, -8.0/12.0, 1.0/12.0};

	gradx.init(curframe.getWidth(), curframe.getHeight());
	grady.init(curframe.getWidth(), curframe.getHeight());
	gradt.init(curframe.getWidth(), curframe.getHeight());
	
	for (int j=gsize/2; j<gradx.getHeight()-gsize/2; j++)
		for (int i=gsize/2; i<gradx.getWidth()-gsize/2; i++)
		{
			gx=0.0;
			gy=0.0;
			for (int k=-gsize/2; k<=gsize/2; k++)
			{
				gx+=curframe.pixel(i+k,j)*D5[k+gsize/2];
				gy+=curframe.pixel(i,j+k)*D5[k+gsize/2];
			}
			
			gradx.pixel(i,j) = gx;
			grady.pixel(i,j) = gy;

			// seperated just in case I use a bigger mask later!
			gt=0.0;
			for (int k=0; k<5; k++)
				gt+=imagein[k].getPixel(i,j)*D5[k];
			//gt=imagein[2].getPixel(i,j)-imagein[1].getPixel(i,j);
			
			gradt.pixel(i,j)=gt;

			
		}
	
	smoothGaussian(gradx,variance);
	smoothGaussian(grady,variance);
	smoothGaussian(gradt,variance);

	// Display gradients in x, y and t directions
	//mm_image<double> tempdisp;
	//tempdisp=gradx;
	//normalise(tempdisp);
	//tempdisp.display();
	//tempdisp.clear();
	//tempdisp=grady;
	//normalise(tempdisp);
	//tempdisp.display();
	//tempdisp.clear();
	//tempdisp=gradt;
	//normalise(tempdisp);
	//tempdisp.display();
		
	xflow.init(gradt.getWidth(),gradt.getHeight());
	yflow.init(gradt.getWidth(),gradt.getHeight());
	
   int
     winSize = 2 * ( windowdist / 2 ) + 1,
     nRows = winSize * winSize;

   ublas::matrix< double, ublas::column_major > A( nRows, 2 );
   ublas::vector< double > b( nRows );

	for (int y=windowdist/2; y<gradt.getHeight()-windowdist/2; y+=windowdist)
		for (int x=windowdist/2; x<gradt.getWidth()-windowdist/2; x+=windowdist)
		{		

			int rowcount=0;
			for (int j=-windowdist/2; j<=windowdist/2; j++)
				for (int i=-windowdist/2; i<=windowdist/2; i++)
				{
					A( rowcount, 0 ) = gradx.pixel(i+x,j+y);
					A( rowcount, 1 ) = grady.pixel(i+x,j+y);
					b( rowcount ) = -gradt.pixel(i+x,j+y);

					rowcount++;

				}

         assert( rowcount == nRows );         

         mm_gels< double > gels;
         ublas::vector< double > temp( gels( A, b ) );

         // mm_inv< double > inv;
         // ublas::vector< double > temp
         //   ( prod( prod( inv( prod( A, trans( A ) ) ), trans( A ) ), b ) );
         
         
			// ensure that motion estimated is less than the windowdist
			//if (tempmat(0,0)>windowdist/2 || tempmat(1,0)>windowdist/2)
			//	continue;

			//cerr << tempmat << endl;
			xflow.pixel(x,y) = temp(0);
			yflow.pixel(x,y) = temp(1);

		}


}







// get the x direction vector flow
void mm_opticflow_diff_tool::getXflow(mm_image<double>& xf)
{
	xf=xflow;
}
	
// get the y direction vector flow
void mm_opticflow_diff_tool::getYflow(mm_image<double>& yf)
{
	yf=yflow;
}

}

#endif
