#include "mm_segment_hashmap_tool.h"

namespace mimas {

mm_segment_hashmap_tool:: mm_segment_hashmap_tool(void)
{
	qlevels=10;
	hashmap=NULL;
	mapCreated=false;
	quantLevelsSet=false;
}


mm_segment_hashmap_tool::~mm_segment_hashmap_tool(void)
{
		if (hashmap)
		{

			for (int l=0; l<qlevels; l++)
			{
				for (int m=0; m<qlevels; m++)
					delete[] hashmap[l][m];
				delete[] hashmap[l];
			}

			delete[] hashmap;
			hashmap=NULL;
		}
	}

void mm_segment_hashmap_tool::setQuantisationLevels(int levels)
{
	if (levels>0)
	{
		if (hashmap)
		{
			for (int l=0; l<qlevels; l++)
			{
				for (int m=0; m<qlevels; m++)
					delete[] hashmap[l][m];
				delete[] hashmap[l];
			}

			delete[] hashmap;
			hashmap=NULL;
		}
			
		qlevels=levels;
		mapCreated=false;

		hashmap=new int**[qlevels];
		for (int l=0; l<qlevels; l++)
		{
			hashmap[l]=new int*[qlevels];
			for (int m=0; m<qlevels; m++)
			{
				hashmap[l][m]=new int[qlevels];
				for (int n=0; n<qlevels; n++)
					hashmap[l][m][n]=0;
			}


		}

		quantLevelsSet=true;
	}
}


void mm_segment_hashmap_tool::storeColours(mm_image< mm_rgba< unsigned char > > &imagein)
{

	if (!quantLevelsSet)
		setQuantisationLevels(qlevels);
	
	int u,v,w;
	double d; // demonimator

	for (int j=0; j<imagein.getHeight(); j++)
		for (int i=0; i<imagein.getWidth(); i++)
		{
         mm_rgba< unsigned char > *pixel = &imagein.pixel( i, j );
			d=(double)pixel->getRed()+pixel->getGreen()+pixel->getBlue()+1;
			u=(int)((double)qlevels*(double)pixel->getRed()/d);
			v=(int)((double)qlevels*(double)pixel->getGreen()/d);
			w=(int)((double)qlevels*(double)pixel->getBlue()/d);

			//cerr << u << " " << v << " " << w << endl;
			hashmap[u][v][w]++;
		}
	
	mapCreated=true;
}

void mm_segment_hashmap_tool::segment( const mm_image< mm_rgba< unsigned char > > &imagein, mm_image< mm_rgba< unsigned char > > &imageout, int thres)
{
	if (!mapCreated) return;

   imageout.init(imagein.getWidth(),imagein.getHeight());

	int u,v,w;
	double d; // demonimator

	for (int j=0; j<imagein.getHeight(); j++)
		for (int i=0; i<imagein.getWidth(); i++)
		{
         const mm_rgba< unsigned char > *pixel = &imagein.pixel( i, j );
			d=(double)pixel->getRed()+pixel->getGreen()+pixel->getBlue()+1;
			u=(int)((double)qlevels*(double)pixel->getRed()/d);
			v=(int)((double)qlevels*(double)pixel->getGreen()/d);
			w=(int)((double)qlevels*(double)pixel->getBlue()/d);

			if (hashmap[u][v][w]>thres)
           imageout.pixel(i,j) = mm_rgba< unsigned char >( 255, 255, 255 );
			else
           imageout.pixel(i,j) = mm_rgba< unsigned char >(   0,   0,   0 );
		}
}

}
