
#include "mm_video4linux.h"

// $Header: /cvs/mimas/lib/mm_video4linux.cc,v 1.9 2004/12/15 19:17:21 engjw Exp $
//
//

namespace mimas {
using namespace std;

mm_video4linux::mm_video4linux(void)
{
	initialised=false;
	setDevice("/dev/video0");

	chan_number=0;
	chan_type=VIDEO_TYPE_CAMERA;
	chan_norm=VIDEO_MODE_PAL;
	
	width=640; 
	height=480;
	palette=VIDEO_PALETTE_RGB24;
	
	depth = 24; /* use 24 for color or 8 for b&w */ 
	brightness = hue = colour = contrast = whiteness = 32767;
	

}


void mm_video4linux::setDevice(char *devname) throw (mm_exception)
{
  MMERROR( !initialised, mm_exception, ,
           "mm_video4linux:: Illegal attempt to reassign device after init" );
	
  strncpy(VIDEO_DEVICE, devname, sizeof(VIDEO_DEVICE));
}

void mm_video4linux::setBrightnessLevel(int b)
{
	brightness=b;
	pic.brightness=brightness;

	if (initialised)
		if (ioctl(fd, VIDIOCSPICT, &pic)  == -1) 
		{
			perror("ioctl VIDIOCSPICT - unable to set brightness");
			return;
		}
}

void mm_video4linux::setHueLevel(int h)
{
	hue=h;
	pic.hue=hue;

	if (initialised)
		if (ioctl(fd, VIDIOCSPICT, &pic)  == -1) 
		{
			perror("ioctl VIDIOCSPICT - unable to set hue");
			return;
		}
}

void mm_video4linux::setColourLevel(int c)
{
	colour=c;
	pic.colour=colour;

	if (initialised)
		if (ioctl(fd, VIDIOCSPICT, &pic)  == -1) 
		{
			perror("ioctl VIDIOCSPICT - unable to set colour");
			return;
		}
}

void mm_video4linux::setContrastLevel(int c)
{
	contrast=c;
	pic.contrast=contrast;

	if (initialised)
		if (ioctl(fd, VIDIOCSPICT, &pic)  == -1) 
		{
			perror("ioctl VIDIOCSPICT - unable to set contrast");
			return;
		}
}

void mm_video4linux::setWhitenessLevel(int w)
{
	whiteness=w;
	pic.whiteness=whiteness;

	if (initialised)
		if (ioctl(fd, VIDIOCSPICT, &pic)  == -1) 
		{
			perror("ioctl VIDIOCSPICT - unable to set whiteness");
			return;
		}
}

void mm_video4linux::setDepth(int d) throw (mm_exception)
{
  MMERROR( !initialised, mm_exception, ,
           "mm_video4linux:: Illegal attempt to set palette after init" );
  depth=d;
}


void mm_video4linux::setPalette(int p) throw (mm_exception)
{
  MMERROR( !initialised, mm_exception, ,
           "mm_video4linux:: Illegal attempt to set palette after init" );
  palette=p;
}


void mm_video4linux::setWidth(int w) throw (mm_exception)
{
  MMERROR( !initialised, mm_exception, ,
           "mm_video4linux:: Illegal attempt to set frame width after init" );
  width=w;
}

void mm_video4linux::setHeight(int h) throw (mm_exception)
{
  MMERROR( !initialised, mm_exception, ,
           "mm_video4linux:: Illegal attempt to set frame height after init" );
  height=h;
}

void mm_video4linux::setChannelNumber(int num) throw (mm_exception)
{
  MMERROR( !initialised, mm_exception, ,
           "mm_video4linux:: Illegal attempt to set channel number after init" );
  chan_number=num;
}


void mm_video4linux::setChannelInputType(int num) throw (mm_exception)
{
  MMERROR( !initialised, mm_exception, ,
           "mm_video4linux:: Illegal attempt to set channel input after init" );
  chan_type=num;
}


void mm_video4linux::setChannelVideoMode(int num) throw (mm_exception)
{
  MMERROR( !initialised, mm_exception, ,
           "mm_video4linux:: Illegal attempt to set channel video mode after init" );
  chan_norm=num;
}


int mm_video4linux::init(void)
{

	// 1. open video device
	fd = open(VIDEO_DEVICE, O_RDWR);


	// 2. get information about video capability
	struct video_capability cap;
	ioctl(fd, VIDIOCGCAP, &cap);

#ifndef NDEBUG
	cerr << "-------------------------------------------" << endl;
	cerr << "Name     : " << cap.name << endl;
	cerr << "Type     : " << cap.type << endl;
	cerr << "Channels : " << cap.channels << endl;
	cerr << "Audios   : " << cap.audios << endl;
	cerr << "Maxwidth : " << cap.maxwidth << endl;
	cerr << "Maxheight: " << cap.maxheight << endl;
	cerr << "Minwidth : " << cap.minwidth << endl;
	cerr << "Minheight: " << cap.minheight << endl;
#endif

	// 3. set channel number for capturing from camera
	struct video_channel chan;
	chan.channel=chan_number;

	/* Get the current settings */
	if (ioctl(fd, VIDIOCGCHAN, &chan) == -1) 
	{
#ifndef NDEBUG
		perror("ioctl VIDIOCGCHAN");
#endif
		//return(-1);
	}

	/* The next line sets no tuners and no audio for the channel */ 
	chan.flags = 0;

	/* The next line sets input type to be of type camera */
	//chan.type = VIDEO_TYPE_TV;
	chan.type=chan_type;

	/* NTSC is an American standard. Some European countries use PAL */
	//chan.norm = VIDEO_MODE_AUTO;
	chan.norm=chan_norm;

	/* Set the new settings */ 
	if (ioctl(fd, VIDIOCSCHAN, &chan) == -1) 
	{
#ifndef NDEBUG
		perror("ioctl VIDIOCSCHAN");
#endif
		//return(-1);
	}

#ifndef NDEBUG
	cerr << "-------------------------------------------" << endl;
	cerr << "Channel : " << chan.channel << endl;
	cerr << "Name    : " << chan.name << endl;
	cerr << "Tuners  : " << chan.tuners << endl;
	cerr << "Type    : " << chan.type << endl;
	cerr << "Norm    : " << chan.norm << endl;
#endif



	// 4. Set the properties of the image window where the image will be captured. 
	struct video_window win;
	ioctl(fd, VIDIOCGWIN, &win);
	win.x = win.y = win.chromakey = 0 ;
	win.width = width;
	win.height = height;
	win.flags = 0;
	if (ioctl(fd, VIDIOCSWIN, &win) == -1) 
	{
		perror("ioctl VIDIOCSWIN");
		return(-1);
	}



	// 5. Set the properties of the image to be captured 
	//struct video_picture pic;
	if (ioctl(fd, VIDIOCGPICT, &pic) == -1)
	{
		perror("ioctl VIDIOCGPICT");
		return(-1);
	}
#ifndef NDEBUG
	cerr << "-------------------------------------------" << endl;
	cerr << "Depth   : " << pic.depth << endl; 
	cerr << "palette : " << pic.palette << endl; 
#endif


	pic.depth=depth;
	pic.palette=palette;
	pic.brightness=brightness;
	pic.hue=hue;
	pic.colour=colour;
	pic.contrast=contrast;
	pic.whiteness=whiteness;

	if (ioctl(fd, VIDIOCSPICT, &pic)  == -1) 
	{
		perror("ioctl VIDIOCSPICT");
		return(-1);
	}

	// 6. Determine size of card memory to map to user space 
	struct video_mbuf m_buf;
	ioctl(fd, VIDIOCGMBUF, &m_buf); 
	/* For simple image grabbing, do not consider the frames and * offsets field of m_buf */ 
	if (ioctl(fd, VIDIOCGMBUF, &m_buf)  == -1) 
	{
		perror("ioctl VIDIOCGMBUF");
		return(-1);
	}
#ifndef NDEBUG
	cerr << "m_buf.size = " << m_buf.size << endl;
#endif

	// 7. Map video capture card memory to user space memory 
	//void *map;
	map = mmap(0, m_buf.size, PROT_READ, MAP_SHARED, fd, 0);
	/* PROT_READ specifies that the memory mapped will only be read from. 
	 * MAP_SHARED specifies that a change to the video memory will change
	 * the user space memory
	 * A good description of mmap() is available at
	 * http://www.opengroup.org/onlinepubs/7908799/xsh/mmap.html */ 

	// 8. Set the capture parameters for the mapped area 
	// struct video_mmap v_map;
	v_map.frame = 0;
	v_map.width = width;
	v_map.height = height;
	v_map.format = palette;

	initialised=true;
   return 0;
}


void mm_video4linux::grab(mm_image< unsigned char > &imagein) throw (mm_exception)
{
   capture();

	// 11. transfer data to mm_image
	int img_h=v_map.height;
	int img_w=v_map.width;

	unsigned char *p;
	p=(unsigned char *)map; 

	unsigned char *rgb;
	rgb=(unsigned char *)malloc(3*img_h*img_w*sizeof(unsigned char));

	//yuv420p_to_rgb(rgb, p, img_w, img_h, 24);
	imagein.init(img_w,img_h);

	//p=rgb;
   for ( int y=0; y<img_h; y++ )
     for ( int x=0; x<img_w; x++ ) { 
       imagein.pixel(x,y) = ((*(p+2))+(*(p+1))+(*(p)))/3;
       p = p + 3; 
       //	p++;
     } 

	free(rgb);

}


void mm_video4linux::grab( mm_image< mm_rgba< unsigned char > > &imagein)
  throw (mm_exception)
{
   capture();

	// 11. transfer data to mm_image
	int img_h=v_map.height;
	int img_w=v_map.width;

	unsigned char *p;
	p=(unsigned char *)map; 

	unsigned char *rgb;
	rgb=(unsigned char *)malloc(3*img_h*img_w*sizeof(unsigned char));

	//yuv420p_to_rgb(rgb, p, img_w, img_h, 24);


   imagein.init(img_w,img_h);

	//p=rgb;
	// int r,g,b;
   for ( int y=0; y<img_h; y++ )
     for ( int x=0; x<img_w; x++ ) { 
       imagein.pixel( x, y ) =
         mm_rgba< unsigned char >( (int)*(p+2), (int)*(p+1), (int)*(p) );
       p = p + 3; 
     } 

	free(rgb);
}



int mm_video4linux::limit(int x)
{
	return ((x)>0xffffff?0xff: ((x)<=0xffff?0:((x)>>16)));
}

void mm_video4linux::copy420Block (int yTL, int yTR, int yBL, int yBR, int u, int v, int rowPixels, unsigned char *rgb, int bits)
{
	const int rvScale=91881;
	const int guScale=-22553;
	const int gvScale=-46801;
	const int buScale=116129;
	const int yScale=65536;
	int r,g,b;

	g=guScale*u+gvScale*v;
	r=rvScale*v;
	b=buScale*u;

	yTL *= yScale;
	yTR *= yScale;
	yBL *= yScale;
	yBR *= yScale;

	if (bits==24)
	{
		/* Write out top two pixels */
		rgb[0]=limit(b+yTL); rgb[1]=limit(g+yTL); rgb[2]=limit(r+yTL);
		rgb[3]=limit(b+yTR); rgb[4]=limit(g+yTR); rgb[5]=limit(r+yTR);
		rgb += 3*rowPixels;

		/* Skip down to next line to write out bottom two pixels */
		rgb[0]=limit(b+yBL); rgb[1]=limit(g+yBL); rgb[2]=limit(r+yBL);
		rgb[3]=limit(b+yBR); rgb[4]=limit(g+yBR); rgb[5]=limit(r+yBR);

	}
	else if (bits==16)
	{
		/* Write out top two pixels */
		rgb[0] = ((limit(b+yTL) >> 3) & 0x1F)
			| ((limit(g+yTL) << 3) & 0xE0);
		rgb[1] = ((limit(g+yTL) >> 5) & 0x07)
			| (limit(r+yTL) & 0xF8);

		rgb[2] = ((limit(b+yTR) >> 3) & 0x1F)
			| ((limit(g+yTR) << 3) & 0xE0);
		rgb[3] = ((limit(g+yTR) >> 5) & 0x07)
			| (limit(r+yTR) & 0xF8);

		/* Skip down to next line to write out bottom two pixels */
		rgb += 2 * rowPixels;

		rgb[0] = ((limit(b+yBL) >> 3) & 0x1F)
			| ((limit(g+yBL) << 3) & 0xE0);
		rgb[1] = ((limit(g+yBL) >> 5) & 0x07)
			| (limit(r+yBL) & 0xF8);

		rgb[2] = ((limit(b+yBR) >> 3) & 0x1F)
			| ((limit(g+yBR) << 3) & 0xE0);
		rgb[3] = ((limit(g+yBR) >> 5) & 0x07)
			| (limit(r+yBR) & 0xF8);
	}
}

int mm_video4linux::yuv420p_to_rgb (unsigned char *out, unsigned char *in, int width, int height, int bits)
{
	const int numpix=width*height;
	const unsigned int bytes = bits >> 3;
	int h, w, y00, y01, y10, y11, u, v;
	unsigned char *pY = in;
	unsigned char *pU = pY + numpix;
	unsigned char *pV = pU + numpix / 4;
	unsigned char *pOut = out;

	if (!out || !in)
		return -1;

	for (h=0; h<=height-2; h+=2)
	{
		for (w=0; w<=width-2; w+=2)
		{
			y00=*(pY);
			y01=*(pY+1);
			y10=*(pY+width);
			y11=*(pY + width +1);
			u=(*pU++) - 128;
			v=(*pV++) - 128;

			copy420Block(y00,y01,y10,y11,u,v,width,pOut,bits);

			pY+=2;
			pOut+=bytes << 1;
		}

		pY +=width;
		pOut +=width*bytes;
	}
	return 0;
}

void mm_video4linux::capture(void) throw (mm_exception)
{
	if (!initialised) init();

	// 9. Capture a frame from the camera 
   MMERROR( ioctl(fd, VIDIOCMCAPTURE, &v_map) != -1, mm_exception, ,
            "ioctl VIDIOCMCAPTURE" );

	// 10. Wait for end of frame
	MMERROR( ioctl(fd, VIDIOCSYNC, &v_map) != -1, mm_exception, ,
            "ioctl VIDIOSYNC" );
}


int mm_video4linux::getWidth(void)
{
	if (!initialised) init();
	return v_map.width;
}

int mm_video4linux::getHeight(void)
{
	if (!initialised) init();
	return v_map.height;
}
}
