#include "mm_config.h"

#ifdef HAVE_XALAN

#ifdef SIMPLEXPATH
#include <xalan/XPath/XPathEvaluator.hpp>
#include <xalan/DOMSupport/XalanDocumentPrefixResolver.hpp>
#else
#include <xalan/XPath/XPath.hpp>
#include <xalan/XPath/ElementPrefixResolverProxy.hpp>
#include <xalan/XPath/XPathProcessorImpl.hpp>
#endif
#include "mm_xml_document.h"
#include "mm_xml_reference.h"
#include "mm_xml_node_reference_list.h"

using namespace boost;

namespace mimas {

mm_xml_reference< XalanNode > mm_xml_reference< XalanElement >::selectNode
  ( const char *xPath ) const throw (mm_exception,XMLException)
{
  mm_xml_node_reference_list nodeList( selectNodes( xPath ) );
  MMERROR( nodeList.getLength() >= 1, mm_exception, ,
           "Could not find XML node with x-path \"" << xPath  << "\"." );
  MMERROR( nodeList.getLength() <= 1, mm_exception, ,
           "There are multiple nodes with x-path \"" << xPath << "\"." );
  assert( nodeList.getLength() == 1 );
  return nodeList.item( 0 );
}

mm_xml_node_reference_list mm_xml_reference< XalanElement >::selectNodes
  ( const char *xPath ) const throw (mm_exception,XMLException)
{
  MMERROR( document != NULL, mm_exception, ,
           "Can not select nodes from empty XML document." );
#ifdef SIMPLEXPATH
  XalanDocumentPrefixResolver thePrefixResolver( document );
  XPathEvaluator theEvaluator;
  NodeRefList obj =
    theEvaluator.selectNodeList( *document->getDOMSupport(), node,
                                 XalanDOMString( xPath ).c_str(),
                                 thePrefixResolver );
#else
  XPathProcessorImpl theXPathProcessor;
  XPath * const contextXPath = document->getXPathFactory()->create();
  theXPathProcessor.initXPath( *contextXPath,
  // With which version was this introduced exactly?
#if _XALAN_VERSION >= 10500
                               *document->getXPathConstructionContext(),
#endif
                               XalanDOMString( xPath ),
                               ElementPrefixResolverProxy
                               ( node, *document->getEnvSupport(),
                                 *document->getDOMSupport() ) );
  shared_ptr< XObjectPtr > obj( new XObjectPtr );
  *obj =
    contextXPath->execute( node,
                           ElementPrefixResolverProxy
                           ( node, *document->getEnvSupport(),
                             *document->getDOMSupport() ),
                           *document->getExecutionContext() );
#endif
  return mm_xml_node_reference_list( obj, document );
}

bool mm_xml_reference< XalanElement >::existNode( const char *xPath ) const
  throw (XMLException)
{
  return selectNodes( xPath ).getLength() >= 1;
}

};

#endif
